#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import inspect
from unittest import mock

from oslo_config import cfg
from oslo_utils import uuidutils

from ironic.common import exception
from ironic.common import network
from ironic.common import states
from ironic.conductor import task_manager
from ironic.drivers.modules.network import (
    ironic_networking as standalone)

from ironic.networking import api as networking_api
from ironic import objects
from ironic.tests.unit.db import base as db_base
from ironic.tests.unit.objects import utils as obj_utils

CONF = cfg.CONF


class IronicNetworkingTestCase(db_base.DbTestCase):

    def setUp(self):
        super(IronicNetworkingTestCase, self).setUp()
        self.config(enabled_hardware_types=['fake-hardware'],
                    enabled_network_interfaces=['ironic-networking'])
        self.node = obj_utils.create_test_node(
            self.context,
            network_interface='ironic-networking')
        self.interface = standalone.IronicNetworking()

    def _generate_test_mac(self):
        rand = uuidutils.generate_uuid().replace('-', '')
        sections = [rand[i:i + 2] for i in range(0, 8, 2)]
        return '52:54:%s:%s:%s:%s' % tuple(sections)

    def _create_test_port(self, node_id=None, **kwargs):
        """Create a test port with default values."""
        port_kwargs = {
            'uuid': uuidutils.generate_uuid(),
            'node_id': node_id or self.node.id,
            'address': kwargs.pop('address', self._generate_test_mac()),
            'local_link_connection': {
                'switch_id': '00:11:22:33:44:55',
                'port_id': 'GigE1/0/1',
                'switch_info': 'test-switch'
            },
            'extra': {}
        }
        port_kwargs.update(kwargs)
        return obj_utils.create_test_port(self.context, **port_kwargs)

    def _create_test_portgroup(self, node_id=None, **kwargs):
        """Create a test portgroup with default values."""
        portgroup_kwargs = {
            'node_id': node_id or self.node.id,
            'name': 'test-portgroup',
            'extra': {}
        }
        portgroup_kwargs.update(kwargs)
        return obj_utils.create_test_portgroup(
            self.context, **portgroup_kwargs)

    def _create_port_copy(self, original_port, **kwargs):
        """Create a copy of a port with modified attributes.

        (not persisted to DB).
        """
        # Create a copy of the port object for testing "current" state
        port_copy = objects.Port(self.context)
        for field in original_port.fields:
            if (hasattr(original_port, field)
                    and original_port.obj_attr_is_set(field)):
                setattr(port_copy, field, getattr(original_port, field))

        # Apply any overrides
        for key, value in kwargs.items():
            setattr(port_copy, key, value)

        return port_copy


class IronicNetworkingValidationTestCase(IronicNetworkingTestCase):

    def test_validate_with_valid_switchport_access(self):
        """Test validation with valid access switchport configuration."""
        port = self._create_test_port(
            extra={
                'switchport': {
                    'mode': 'access',
                    'native_vlan': 100
                }
            }
        )
        with task_manager.acquire(self.context, self.node.uuid) as task:
            task.ports = [port]
            task.portgroups = []
            self.interface.validate(task)

    def test_validate_with_valid_switchport_trunk(self):
        """Test validation with valid trunk switchport configuration."""
        port = self._create_test_port(
            pxe_enabled=False,
            extra={
                'switchport': {
                    'mode': 'trunk',
                    'native_vlan': 100,
                    'allowed_vlans': [100, 200, 300]
                }
            }
        )
        with task_manager.acquire(self.context, self.node.uuid) as task:
            task.ports = [port]
            task.portgroups = []
            self.interface.validate(task)

    def test_validate_with_valid_switchport_hybrid(self):
        """Test validation with valid hybrid switchport configuration."""
        port = self._create_test_port(
            pxe_enabled=False,
            extra={
                'switchport': {
                    'mode': 'hybrid',
                    'native_vlan': 100,
                    'allowed_vlans': [100, 200, 300]
                }
            }
        )
        with task_manager.acquire(self.context, self.node.uuid) as task:
            task.ports = [port]
            task.portgroups = []
            self.interface.validate(task)

    def test_validate_with_invalid_switchport_mode(self):
        """Test validation with invalid switchport mode."""
        port = self._create_test_port(
            extra={
                'switchport': {
                    'mode': 'invalid',
                    'native_vlan': 100
                }
            }
        )
        with task_manager.acquire(self.context, self.node.uuid) as task:
            task.ports = [port]
            task.portgroups = []
            self.assertRaises(exception.InvalidParameterValue,
                              self.interface.validate, task)

    def test_validate_with_invalid_lag_schema(self):
        """Portgroup validation fails when schema invalid."""
        portgroup = self._create_test_portgroup(
            extra={'lag': {'mode': 'access'}}
        )
        with task_manager.acquire(self.context, self.node.uuid) as task:
            task.ports = []
            task.portgroups = [portgroup]
            self.assertRaises(exception.InvalidParameterValue,
                              self.interface.validate, task)

    def test_validate_portgroup_no_member_ports(self):
        """Portgroup validation fails when members missing."""
        portgroup = self._create_test_portgroup(
            extra={'lag': {
                'mode': 'access',
                'aggregation_mode': 'static'
            }}
        )
        with task_manager.acquire(self.context, self.node.uuid) as task:
            task.ports = []
            task.portgroups = [portgroup]
            self.assertRaises(exception.InvalidParameterValue,
                              self.interface.validate, task)

    def test_validate_switchport_missing_switch_id(self):
        """Port validation fails when switch_id missing."""
        port = self._create_test_port(
            local_link_connection={'port_id': 'GigE1/0/1'},
            extra={'switchport': {'mode': 'access', 'native_vlan': 100}}
        )
        with task_manager.acquire(self.context, self.node.uuid) as task:
            task.ports = [port]
            task.portgroups = []
            self.assertRaises(exception.InvalidParameterValue,
                              self.interface.validate, task)

    def test_validate_switchport_missing_port_id(self):
        """Port validation fails when port_id missing."""
        port = self._create_test_port(
            local_link_connection={'switch_id': '00:11:22:33:44:55'},
            extra={'switchport': {'mode': 'access', 'native_vlan': 100}}
        )
        with task_manager.acquire(self.context, self.node.uuid) as task:
            task.ports = [port]
            task.portgroups = []
            self.assertRaises(exception.InvalidParameterValue,
                              self.interface.validate, task)

    def test_validate_pxe_enabled_port_with_access_mode(self):
        """Test validation passes for PXE-enabled port with access mode."""
        port = self._create_test_port(
            pxe_enabled=True,
            extra={
                'switchport': {
                    'mode': 'access',
                    'native_vlan': 100
                }
            }
        )
        with task_manager.acquire(self.context, self.node.uuid) as task:
            task.ports = [port]
            task.portgroups = []
            self.interface.validate(task)

    def test_validate_pxe_enabled_port_with_trunk_mode_fails(self):
        """Test validation fails for PXE-enabled port with trunk mode."""
        port = self._create_test_port(
            pxe_enabled=True,
            extra={
                'switchport': {
                    'mode': 'trunk',
                    'native_vlan': 100,
                    'allowed_vlans': [100, 200, 300]
                }
            }
        )
        with task_manager.acquire(self.context, self.node.uuid) as task:
            task.ports = [port]
            task.portgroups = []
            self.assertRaises(exception.InvalidParameterValue,
                              self.interface.validate, task)

    def test_validate_pxe_enabled_port_with_hybrid_mode_fails(self):
        """Test validation fails for PXE-enabled port with hybrid mode."""
        port = self._create_test_port(
            pxe_enabled=True,
            extra={
                'switchport': {
                    'mode': 'hybrid',
                    'native_vlan': 100,
                    'allowed_vlans': [100, 200, 300]
                }
            }
        )
        with task_manager.acquire(self.context, self.node.uuid) as task:
            task.ports = [port]
            task.portgroups = []
            self.assertRaises(exception.InvalidParameterValue,
                              self.interface.validate, task)

    def test_validate_pxe_disabled_port_with_trunk_mode(self):
        """Test validation passes for non-PXE port with trunk mode."""
        port = self._create_test_port(
            pxe_enabled=False,
            extra={
                'switchport': {
                    'mode': 'trunk',
                    'native_vlan': 100,
                    'allowed_vlans': [100, 200, 300]
                }
            }
        )
        with task_manager.acquire(self.context, self.node.uuid) as task:
            task.ports = [port]
            task.portgroups = []
            self.interface.validate(task)

    def test_validate_pxe_disabled_port_with_hybrid_mode(self):
        """Test validation passes for non-PXE port with hybrid mode."""
        port = self._create_test_port(
            pxe_enabled=False,
            extra={
                'switchport': {
                    'mode': 'hybrid',
                    'native_vlan': 100,
                    'allowed_vlans': [100, 200, 300]
                }
            }
        )
        with task_manager.acquire(self.context, self.node.uuid) as task:
            task.ports = [port]
            task.portgroups = []
            self.interface.validate(task)

    def test_validate_pxe_enabled_port_without_switchport(self):
        """Test validation passes for PXE-enabled port without switchport."""
        port = self._create_test_port(
            pxe_enabled=True,
            extra={}
        )
        with task_manager.acquire(self.context, self.node.uuid) as task:
            task.ports = [port]
            task.portgroups = []
            self.interface.validate(task)

    def test_validate_switchport_missing_local_link_outside_inspection(self):
        """Port validation fails when not in inspection state.

        When a port has switchport configuration but is missing
        local_link_connection and the node is not in an inspection state,
        validation should raise an error since we cannot determine which
        physical switchport to configure.
        """
        # Set node to a non-inspection state (e.g., MANAGEABLE)
        self.node.provision_state = states.MANAGEABLE
        self.node.save()

        port = self._create_test_port(
            local_link_connection=None,
            extra={'switchport': {'mode': 'access', 'native_vlan': 100}}
        )
        with task_manager.acquire(self.context, self.node.uuid) as task:
            task.ports = [port]
            task.portgroups = []
            self.assertRaises(exception.InvalidParameterValue,
                              self.interface.validate, task)

    def test_validate_switchport_missing_local_link_during_inspection(self):
        """Port validation passes when in inspection state.

        When a port has switchport configuration but is missing
        local_link_connection while the node is in an inspection state,
        validation should pass with a debug log since local_link_connection
        may be populated during inspection.
        """
        # Set node to an inspection state
        self.node.provision_state = states.INSPECTING
        self.node.save()

        port = self._create_test_port(
            local_link_connection=None,
            extra={'switchport': {'mode': 'access', 'native_vlan': 100}}
        )
        with task_manager.acquire(self.context, self.node.uuid) as task:
            task.ports = [port]
            task.portgroups = []
            # Should not raise an exception
            self.interface.validate(task)


class IronicNetworkingPortChangedTestCase(
        IronicNetworkingTestCase):

    def setUp(self):
        super(IronicNetworkingPortChangedTestCase, self).setUp()
        # Ensure node is in a state where a network should be active so that
        # port_changed triggers networking API calls.
        self.node.provision_state = states.DEPLOYING
        self.node.save()

    @mock.patch.object(networking_api, 'reset_port', autospec=True)
    def test_port_changed_deleted_case(self, mock_reset_port):
        """Test port_changed when switchport config is deleted."""
        mock_reset_port.return_value = {'status': 'reset'}

        # Create port with switchport config
        original_port = self._create_test_port(
            extra={
                'switchport': {
                    'mode': 'access',
                    'native_vlan': 100
                }
            }
        )

        # Current port without switchport config
        current_port = self._create_port_copy(original_port, extra={})

        with task_manager.acquire(self.context, self.node.uuid) as task:
            task.ports = [original_port]
            self.interface.port_changed(task, current_port)

        mock_reset_port.assert_called_once_with(
            task.context, '00:11:22:33:44:55', 'GigE1/0/1', 100,
            allowed_vlans=None, default_vlan=None)

    @mock.patch.object(networking_api, 'reset_port', autospec=True)
    def test_port_changed_deleted_skips_without_active_network(
            self, mock_reset_port):
        """Deletion skips when no active network should be configured."""
        self.node.provision_state = states.ENROLL
        self.node.save()
        mock_reset_port.return_value = {'status': 'reset'}
        original_port = self._create_test_port(
            extra={'switchport': {'mode': 'access', 'native_vlan': 100}}
        )
        current_port = self._create_port_copy(original_port, extra={})
        with task_manager.acquire(self.context, self.node.uuid) as task:
            task.ports = [original_port]
            self.interface.port_changed(task, current_port)
        mock_reset_port.assert_not_called()

    @mock.patch.object(networking_api, 'reset_port', autospec=True)
    def test_port_changed_deleted_case_trunk_mode(self, mock_reset_port):
        """Test port_changed when trunk switchport config is deleted."""
        mock_reset_port.return_value = {'status': 'reset'}

        # Create port with trunk switchport config
        original_port = self._create_test_port(
            extra={
                'switchport': {
                    'mode': 'trunk',
                    'native_vlan': 100,
                    'allowed_vlans': [100, 200, 300]
                }
            }
        )

        # Current port without switchport config
        current_port = self._create_port_copy(original_port, extra={})

        with task_manager.acquire(self.context, self.node.uuid) as task:
            task.ports = [original_port]
            self.interface.port_changed(task, current_port)

        mock_reset_port.assert_called_once_with(
            task.context, '00:11:22:33:44:55', 'GigE1/0/1', 100,
            allowed_vlans=[100, 200, 300], default_vlan=None)

    @mock.patch.object(networking_api, 'reset_port', autospec=True)
    def test_port_changed_deleted_case_hybrid_mode(self, mock_reset_port):
        """Test port_changed when hybrid switchport config is deleted."""
        mock_reset_port.return_value = {'status': 'reset'}

        # Create port with hybrid switchport config
        original_port = self._create_test_port(
            extra={
                'switchport': {
                    'mode': 'hybrid',
                    'native_vlan': 100,
                    'allowed_vlans': [100, 200, 300]
                }
            }
        )

        # Current port without switchport config
        current_port = self._create_port_copy(original_port, extra={})

        with task_manager.acquire(self.context, self.node.uuid) as task:
            task.ports = [original_port]
            self.interface.port_changed(task, current_port)

        mock_reset_port.assert_called_once_with(
            task.context, '00:11:22:33:44:55', 'GigE1/0/1', 100,
            allowed_vlans=[100, 200, 300], default_vlan=None)

    @mock.patch.object(networking_api, 'update_port', autospec=True)
    def test_port_changed_created_case_access(self, mock_update_port):
        """Test port_changed when switchport config is created.

        (access mode).
        """
        mock_update_port.return_value = {'status': 'updated'}

        # Original port without switchport config
        original_port = self._create_test_port(extra={})

        # Current port with access switchport config
        current_port = self._create_port_copy(
            original_port,
            extra={
                'switchport': {
                    'mode': 'access',
                    'native_vlan': 100
                }
            }
        )

        with task_manager.acquire(self.context, self.node.uuid) as task:
            task.ports = [original_port]
            self.interface.port_changed(task, current_port)

        mock_update_port.assert_called_once_with(
            task.context, '00:11:22:33:44:55', 'GigE1/0/1',
            'Ironic Port ' + str(current_port.uuid), 'access', 100,
            allowed_vlans=None, lag_name=None, default_vlan=None)

    @mock.patch.object(networking_api, 'update_port', autospec=True)
    def test_port_changed_switchport_added_no_local_link(
            self, mock_update_port):
        """Switchport addition skipped if local_link missing."""
        original_port = self._create_test_port(
            local_link_connection=None,
            extra={}
        )
        current_port = self._create_port_copy(
            original_port,
            extra={'switchport': {'mode': 'access', 'native_vlan': 100}}
        )
        with task_manager.acquire(self.context, self.node.uuid) as task:
            task.ports = [original_port]
            self.interface.port_changed(task, current_port)
        mock_update_port.assert_not_called()

    @mock.patch.object(networking_api, 'update_port', autospec=True)
    def test_port_changed_created_case_trunk(self, mock_update_port):
        """Test port_changed when switchport config is created (trunk mode)."""
        mock_update_port.return_value = {'status': 'updated'}

        # Original port without switchport config
        original_port = self._create_test_port(pxe_enabled=False, extra={})

        # Current port with trunk switchport config
        current_port = self._create_port_copy(
            original_port,
            extra={
                'switchport': {
                    'mode': 'trunk',
                    'native_vlan': 100,
                    'allowed_vlans': [100, 200, 300],
                    'lag_name': 'port-channel1'
                }
            }
        )

        with task_manager.acquire(self.context, self.node.uuid) as task:
            task.ports = [original_port]
            self.interface.port_changed(task, current_port)

        mock_update_port.assert_called_once_with(
            task.context, '00:11:22:33:44:55', 'GigE1/0/1',
            'Ironic Port ' + str(current_port.uuid), 'trunk', 100,
            allowed_vlans=[100, 200, 300],
            lag_name='port-channel1',
            default_vlan=None)

    @mock.patch.object(networking_api, 'update_port', autospec=True)
    def test_port_changed_switchport_added_uses_network_override(
            self, mock_update_port):
        """Network override from config takes precedence over switchport."""
        mock_update_port.return_value = {'status': 'updated'}
        original_port = self._create_test_port(extra={})
        current_port = self._create_port_copy(
            original_port,
            extra={
                'switchport': {
                    'mode': 'access',
                    'native_vlan': 111
                }
            }
        )
        with task_manager.acquire(self.context, self.node.uuid) as task:
            task.ports = [original_port]
            self.config(provisioning_network='access/native_vlan=222',
                        group='ironic_networking')
            self.interface.port_changed(task, current_port)
        mock_update_port.assert_called_once_with(
            task.context, '00:11:22:33:44:55', 'GigE1/0/1',
            'Ironic Port ' + str(current_port.uuid), 'access', 222,
            allowed_vlans=None, lag_name=None, default_vlan=None)

    @mock.patch.object(networking_api, 'update_port', autospec=True)
    def test_port_changed_created_case_hybrid(self, mock_update_port):
        """Test port_changed when switchport config is created.

        (hybrid mode).
        """
        mock_update_port.return_value = {'status': 'updated'}

        # Original port without switchport config
        original_port = self._create_test_port(pxe_enabled=False, extra={})

        # Current port with hybrid switchport config
        current_port = self._create_port_copy(
            original_port,
            extra={
                'switchport': {
                    'mode': 'hybrid',
                    'native_vlan': 100,
                    'allowed_vlans': [100, 200, 300],
                    'lag_name': 'port-channel1'
                }
            }
        )

        with task_manager.acquire(self.context, self.node.uuid) as task:
            task.ports = [original_port]
            self.interface.port_changed(task, current_port)

        mock_update_port.assert_called_once_with(
            task.context, '00:11:22:33:44:55', 'GigE1/0/1',
            'Ironic Port ' + str(current_port.uuid), 'hybrid', 100,
            allowed_vlans=[100, 200, 300],
            lag_name='port-channel1',
            default_vlan=None)

    @mock.patch.object(networking_api, 'update_port', autospec=True)
    def test_port_changed_updated_case(self, mock_update_port):
        """Test port_changed when switchport config is updated."""
        mock_update_port.return_value = {'status': 'updated'}

        # Original port with access switchport config
        original_port = self._create_test_port(
            extra={
                'switchport': {
                    'mode': 'access',
                    'native_vlan': 100
                }
            }
        )

        # Current port with different access switchport config
        current_port = self._create_port_copy(
            original_port,
            extra={
                'switchport': {
                    'mode': 'access',
                    'native_vlan': 200
                }
            }
        )

        with task_manager.acquire(self.context, self.node.uuid) as task:
            task.ports = [original_port]
            self.interface.port_changed(task, current_port)

        mock_update_port.assert_called_once_with(
            task.context, '00:11:22:33:44:55', 'GigE1/0/1',
            'Ironic Port ' + str(current_port.uuid), 'access', 200,
            allowed_vlans=None, lag_name=None, default_vlan=None)

    def test_port_changed_no_change(self):
        """Test port_changed when no switchport config changes."""
        # Same switchport config in both original and current
        switchport_config = {
            'mode': 'access',
            'native_vlan': 100
        }

        original_port = self._create_test_port(
            extra={'switchport': switchport_config}
        )

        current_port = self._create_port_copy(
            original_port,
            extra={'switchport': switchport_config.copy()}
        )

        with task_manager.acquire(self.context, self.node.uuid) as task:
            task.ports = [original_port]
            # Should not raise any exceptions and should not call
            # networking API
            self.interface.port_changed(task, current_port)

    @mock.patch.object(networking_api, 'update_port', autospec=True)
    @mock.patch.object(networking_api, 'reset_port', autospec=True)
    def test_port_changed_handles_local_link_change(
            self, mock_reset_port, mock_update_port):
        """Local link change triggers reset then update."""
        mock_reset_port.return_value = {'status': 'reset'}
        mock_update_port.return_value = {'status': 'updated'}
        original_port = self._create_test_port(
            extra={'switchport': {'mode': 'access', 'native_vlan': 100}}
        )
        current_port = self._create_port_copy(
            original_port,
            local_link_connection={
                'switch_id': '00:11:22:33:44:66',
                'port_id': 'GigE1/0/2'
            }
        )
        with task_manager.acquire(self.context, self.node.uuid) as task:
            task.ports = [original_port]
            self.interface.port_changed(task, current_port)
        mock_reset_port.assert_called_once_with(
            task.context, '00:11:22:33:44:55', 'GigE1/0/1', 100,
            allowed_vlans=None, default_vlan=None)
        mock_update_port.assert_called_once_with(
            task.context, '00:11:22:33:44:66', 'GigE1/0/2',
            'Ironic Port ' + str(current_port.uuid), 'access', 100,
            allowed_vlans=None, lag_name=None, default_vlan=None)

    @mock.patch.object(networking_api, 'reset_port', autospec=True)
    def test_port_changed_deleted_missing_required_fields(self,
                                                          mock_reset_port):
        """Test port_changed deletion case with missing required fields."""
        # Create port with switchport config but missing local_link_connection
        original_port = self._create_test_port(
            # Empty instead of None to pass validation
            local_link_connection={},
            extra={
                'switchport': {
                    'mode': 'access',
                    'native_vlan': 100
                }
            }
        )

        # Current port without switchport config
        current_port = self._create_port_copy(original_port, extra={})

        with task_manager.acquire(self.context, self.node.uuid) as task:
            task.ports = [original_port]
            self.interface.port_changed(task, current_port)

        # Should not call reset_port due to missing fields
        mock_reset_port.assert_not_called()

    @mock.patch.object(networking_api, 'update_port', autospec=True)
    def test_port_changed_created_missing_required_fields(self,
                                                          mock_update_port):
        """Test port_changed creation case with missing required fields."""
        # Original port without switchport config
        original_port = self._create_test_port(extra={})

        # Current port with switchport config but missing native_vlan
        current_port = self._create_port_copy(
            original_port,
            extra={
                'switchport': {
                    'mode': 'access'
                    # Missing native_vlan
                }
            }
        )

        with task_manager.acquire(self.context, self.node.uuid) as task:
            task.ports = [original_port]
            self.assertRaises(
                exception.InvalidParameterValue,
                self.interface.port_changed, task, current_port)

        # Should not call update_port due to invalid config
        mock_update_port.assert_not_called()


class IronicNetworkingPortgroupChangedTestCase(
        IronicNetworkingTestCase):

    def test_portgroup_changed_logs_message(self):
        """Test that portgroup_changed logs an appropriate message."""
        portgroup = self._create_test_portgroup(name='test-pg')

        with task_manager.acquire(self.context, self.node.uuid) as task:
            with mock.patch('ironic.drivers.modules.network.'
                            'ironic_networking.LOG',
                            autospec=True) as mock_log:
                self.interface.portgroup_changed(task, portgroup)
                mock_log.debug.assert_called_once_with(
                    "Portgroup %(portgroup)s (%(name)s) configuration "
                    "changed - portgroup changes not currently supported by "
                    "ironic-networking interface",
                    {'portgroup': portgroup.uuid, 'name': 'test-pg'})

    def test_portgroup_changed_logs_message_unnamed(self):
        """Test that portgroup_changed logs an appropriate message.

        For unnamed portgroup.
        """
        portgroup = self._create_test_portgroup(name=None)

        with task_manager.acquire(self.context, self.node.uuid) as task:
            with mock.patch('ironic.drivers.modules.network.'
                            'ironic_networking.LOG',
                            autospec=True) as mock_log:
                self.interface.portgroup_changed(task, portgroup)
                mock_log.debug.assert_called_once_with(
                    "Portgroup %(portgroup)s (%(name)s) configuration "
                    "changed - portgroup changes not currently supported by "
                    "ironic-networking interface",
                    {'portgroup': portgroup.uuid, 'name': 'unnamed'})


class IronicNetworkingHelperMethodsTestCase(
        IronicNetworkingTestCase):

    def test_get_port_switch_info(self):
        """Test _get_port_switch_info method."""
        port = self._create_test_port()
        switch_id, port_name = self.interface._get_port_switch_info(port)
        self.assertEqual('00:11:22:33:44:55', switch_id)
        self.assertEqual('GigE1/0/1', port_name)

    def test_get_port_switch_info_no_local_link_connection(self):
        """Test _get_port_switch_info with no local_link_connection."""
        port = self._create_test_port(local_link_connection=None)
        switch_id, port_name = self.interface._get_port_switch_info(port)
        self.assertIsNone(switch_id)
        self.assertIsNone(port_name)

    def test_get_port_description(self):
        """Test _get_port_description method."""
        port = self._create_test_port()
        description = self.interface._get_port_description(port)
        self.assertEqual(f'Ironic Port {port.uuid}', description)

    def test_get_portgroup_description_named(self):
        """Test _get_portgroup_description for a named portgroup."""
        portgroup = self._create_test_portgroup(name='pg-name')
        description = self.interface._get_portgroup_description(portgroup)
        self.assertEqual(f'Ironic PortGroup {portgroup.uuid}', description)

    def test_get_portgroup_description_unnamed(self):
        """Test _get_portgroup_description fallback."""
        portgroup = self._create_test_portgroup(name=None)
        description = self.interface._get_portgroup_description(portgroup)
        expected = f'Ironic PortGroup {portgroup.uuid}'
        self.assertEqual(expected, description)

    def test_get_portgroup_lag_name_named(self):
        """Test _get_portgroup_lag_name with explicit name."""
        portgroup = self._create_test_portgroup(name='named-group')
        name = self.interface._get_portgroup_lag_name(portgroup)
        self.assertEqual('named-group', name)

    def test_get_portgroup_lag_name_unnamed(self):
        """Test _get_portgroup_lag_name without name."""
        portgroup = self._create_test_portgroup(name=None)
        name = self.interface._get_portgroup_lag_name(portgroup)
        expected = f'lag-{portgroup.uuid[:8]}'
        self.assertEqual(expected, name)

    def test_get_portgroup_switch_ids(self):
        """Test _get_portgroup_switch_ids collects unique IDs."""
        portgroup = self._create_test_portgroup(
            extra={'lag': {
                'mode': 'access',
                'aggregation_mode': 'static'
            }}
        )
        port_one = self._create_test_port(
            portgroup_id=portgroup.id,
            local_link_connection={
                'switch_id': 'switch-1',
                'port_id': 'Eth1/1'
            },
            extra={'switchport': {'mode': 'access', 'native_vlan': 100}}
        )
        port_two = self._create_test_port(
            portgroup_id=portgroup.id,
            local_link_connection={
                'switch_id': 'switch-2',
                'port_id': 'Eth1/2'
            },
            extra={'switchport': {'mode': 'access', 'native_vlan': 100}}
        )
        with task_manager.acquire(self.context, self.node.uuid) as task:
            task.portgroups = [portgroup]
            task.ports = [port_one, port_two]
            result = self.interface._get_portgroup_switch_ids(
                task, portgroup)
        self.assertCountEqual(['switch-1', 'switch-2'], result)

    def test_get_portgroup_switch_ids_no_lag(self):
        """Test _get_portgroup_switch_ids returns None when unused."""
        portgroup = self._create_test_portgroup(extra={})
        with task_manager.acquire(self.context, self.node.uuid) as task:
            task.portgroups = [portgroup]
            task.ports = []
            result = self.interface._get_portgroup_switch_ids(
                task, portgroup)
        self.assertIsNone(result)

    def test_get_network_type_for_state_mappings(self):
        """Test get_network_type_for_state state mappings."""
        state_map = {
            states.INSPECTING: 'inspection',
            states.INSPECTWAIT: 'inspection',
            states.CLEANING: 'cleaning',
            states.CLEANWAIT: 'cleaning',
            states.DEPLOYING: 'provisioning',
            states.DEPLOYWAIT: 'provisioning',
            states.ACTIVE: 'tenant',
            states.RESCUE: 'rescuing',
            states.RESCUING: 'rescuing',
            states.RESCUEWAIT: 'rescuing',
            states.SERVICING: 'servicing',
            states.SERVICEWAIT: 'servicing'
        }
        for provision_state, expected in state_map.items():
            self.assertEqual(
                expected,
                network.get_network_type_for_state(provision_state))

    def test_get_network_type_for_state_default_idle(self):
        """Test get_network_type_for_state returns idle for unknown states."""
        self.assertEqual(
            'idle', network.get_network_type_for_state(states.MANAGEABLE))
        self.assertEqual(
            'idle', network.get_network_type_for_state(states.ENROLL))

    def test_get_switch_port_config_driver_info(self):
        """Driver info overrides networking configuration."""
        driver_info = dict(self.node.driver_info or {})
        driver_info['provisioning_network'] = (
            'access/native_vlan=66')
        self.node.driver_info = driver_info
        self.node.save()
        with task_manager.acquire(self.context, self.node.uuid) as task:
            config = self.interface._get_switch_port_config(
                task, 'provisioning')
        self.assertEqual('access', config.mode)
        self.assertEqual(66, config.native_vlan)
        self.assertIsNone(config.allowed_vlans)

    def test_get_switch_port_config_conf_fallback(self):
        """Global networking configuration is used when no override."""
        driver_info = dict(self.node.driver_info or {})
        driver_info.pop('cleaning_network', None)
        self.node.driver_info = driver_info
        self.node.save()
        self.config(cleaning_network='trunk/native_vlan=77',
                    group='ironic_networking')
        with task_manager.acquire(self.context, self.node.uuid) as task:
            config = self.interface._get_switch_port_config(
                task, 'cleaning')
        self.assertEqual('trunk', config.mode)
        self.assertEqual(77, config.native_vlan)
        self.assertIsNone(config.allowed_vlans)

    def test_get_switch_port_config_invalid_format(self):
        """Invalid configuration raises InvalidParameterValue."""
        driver_info = dict(self.node.driver_info or {})
        driver_info['provisioning_network'] = 'invalid'
        self.node.driver_info = driver_info
        self.node.save()
        with task_manager.acquire(self.context, self.node.uuid) as task:
            self.assertRaises(exception.InvalidParameterValue,
                              self.interface._get_switch_port_config,
                              task, 'provisioning')

    def test_get_switch_port_config_not_set(self):
        """Returns None when no network is configured."""
        driver_info = dict(self.node.driver_info or {})
        driver_info.pop('provisioning_network', None)
        self.node.driver_info = driver_info
        self.node.save()
        with task_manager.acquire(self.context, self.node.uuid) as task:
            config = self.interface._get_switch_port_config(
                task, 'provisioning')
        self.assertIsNone(config)

    def test_resolve_network_configuration_prefers_network(self):
        """_resolve_network_configuration prefers global network config."""
        driver_info = dict(self.node.driver_info or {})
        driver_info['provisioning_network'] = (
            'hybrid/native_vlan=88/allowed_vlans=88,89')
        self.node.driver_info = driver_info
        self.node.save()
        port = self._create_test_port(
            extra={'switchport': {
                'mode': 'access',
                'native_vlan': 10,
                'allowed_vlans': [10, 11]
            }}
        )
        with (task_manager.acquire(self.context, self.node.uuid) as task):
            task.ports = [port]
            config = self.interface._resolve_network_configuration(
                task, port, 'provisioning')
        self.assertEqual('hybrid', config.mode)
        self.assertEqual(88, config.native_vlan)
        self.assertEqual([88, 89], config.allowed_vlans)

    def test_resolve_network_configuration_fallback_switchport(self):
        """_resolve_network_configuration falls back to switchport."""
        driver_info = dict(self.node.driver_info or {})
        driver_info.pop('provisioning_network', None)
        self.node.driver_info = driver_info
        self.node.save()
        port = self._create_test_port(
            extra={'switchport': {
                'mode': 'hybrid',
                'native_vlan': 55,
                'allowed_vlans': [55, 56]
            }}
        )
        with task_manager.acquire(self.context, self.node.uuid) as task:
            task.ports = [port]
            config = self.interface._resolve_network_configuration(
                task, port, 'provisioning')
        self.assertEqual('hybrid', config.mode)
        self.assertEqual(55, config.native_vlan)
        self.assertEqual([55, 56], config.allowed_vlans)

    def test_validate_portgroup_members_missing_switchport(self):
        """Portgroup validation raises when member lacks switchport."""
        portgroup = self._create_test_portgroup(
            extra={'lag': {
                'mode': 'access',
                'aggregation_mode': 'static'
            }}
        )
        member_port = self._create_test_port(
            portgroup_id=portgroup.id,
            extra={}
        )
        with task_manager.acquire(self.context, self.node.uuid) as task:
            task.portgroups = [portgroup]
            task.ports = [member_port]
            self.assertRaises(exception.InvalidParameterValue,
                              self.interface.validate, task)


class IronicNetworkingNetworkActionsTestCase(
        IronicNetworkingTestCase):

    def setUp(self):
        super(IronicNetworkingNetworkActionsTestCase, self).setUp()
        self.port = self._create_test_port(
            extra={'switchport': {'mode': 'access', 'native_vlan': 100}}
        )

    def _expected_update_call(self, description, mode, native_vlan,
                              allowed_vlans=None, default_vlan=None):
        signature = inspect.signature(networking_api.update_port)
        expected_kwargs = {'allowed_vlans': allowed_vlans}
        if 'lag_name' in signature.parameters:
            expected_kwargs['lag_name'] = None
        if 'default_vlan' in signature.parameters:
            expected_kwargs['default_vlan'] = default_vlan
        return (mock.ANY, '00:11:22:33:44:55', 'GigE1/0/1', description,
                mode, native_vlan), expected_kwargs

    def _prepare_task(self, provision_state):
        self.node.provision_state = provision_state
        self.node.save()
        task = task_manager.acquire(self.context, self.node.uuid)
        task.ports = [self.port]
        return task

    def _configure_driver_info(self, network_type, value):
        key = f'{network_type}_network'
        driver_info = dict(self.node.driver_info or {})
        driver_info[key] = value
        self.node.driver_info = driver_info
        self.node.save()

    def _test_add_network(self, method_name, provision_state,
                          network_type, config_value=None):
        if config_value is not None:
            self._configure_driver_info(network_type, config_value)
        with mock.patch.object(networking_api, 'update_port',
                               autospec=True) as update_mock:
            update_mock.return_value = {'status': 'updated'}
            with self._prepare_task(provision_state) as task:
                getattr(self.interface, method_name)(task)
        update_mock.assert_called()
        return update_mock

    def _test_remove_network(self, method_name, provision_state,
                             network_type, config_value=None):
        if config_value is not None:
            self._configure_driver_info(network_type, config_value)
        with mock.patch.object(networking_api, 'reset_port',
                               autospec=True) as reset_mock:
            reset_mock.return_value = {'status': 'reset'}
            with self._prepare_task(provision_state) as task:
                getattr(self.interface, method_name)(task)
        reset_mock.assert_called()
        return reset_mock

    def test_add_provisioning_network_with_config(self):
        mock_call = self._test_add_network(
            'add_provisioning_network', states.DEPLOYING, 'provisioning',
            'access/native_vlan=200')
        expected_args, expected_kwargs = self._expected_update_call(
            f'Ironic Port {self.port.uuid}', 'access', 200)
        mock_call.assert_called_with(*expected_args, **expected_kwargs)

    def test_remove_provisioning_network_with_config(self):
        mock_call = self._test_remove_network(
            'remove_provisioning_network', states.DEPLOYING,
            'provisioning', 'access/native_vlan=200')
        mock_call.assert_called_with(
            mock.ANY, '00:11:22:33:44:55', 'GigE1/0/1', 200,
            allowed_vlans=None, default_vlan=None)

    def test_add_provisioning_network_with_idle(self):
        CONF.set_override('idle_network', 'access/native_vlan=1',
                          group='ironic_networking')
        mock_call = self._test_add_network(
            'add_provisioning_network', states.DEPLOYING, 'provisioning',
            'access/native_vlan=200')
        expected_args, expected_kwargs = self._expected_update_call(
            f'Ironic Port {self.port.uuid}',
            'access', 200, default_vlan=1)
        mock_call.assert_called_with(*expected_args, **expected_kwargs)

    def test_remove_provisioning_network_with_idle(self):
        CONF.set_override('idle_network', 'access/native_vlan=1',
                          group='ironic_networking')
        mock_call = self._test_remove_network(
            'remove_provisioning_network', states.DEPLOYING,
            'provisioning', 'access/native_vlan=200')
        mock_call.assert_called_with(
            mock.ANY, '00:11:22:33:44:55', 'GigE1/0/1', 200,
            allowed_vlans=None, default_vlan=1)

    def test_add_cleaning_network_without_config(self):
        mock_call = self._test_add_network('add_cleaning_network',
                                           states.CLEANING, 'cleaning')
        expected_args, expected_kwargs = self._expected_update_call(
            f'Ironic Port {self.port.uuid}', 'access', 100)
        mock_call.assert_called_with(*expected_args, **expected_kwargs)

    def test_remove_cleaning_network_without_config(self):
        mock_call = self._test_remove_network('remove_cleaning_network',
                                              states.CLEANING, 'cleaning')
        mock_call.assert_called_with(
            mock.ANY, '00:11:22:33:44:55', 'GigE1/0/1', 100,
            allowed_vlans=None, default_vlan=None)

    def test_add_rescuing_network_with_config(self):
        mock_call = self._test_add_network(
            'add_rescuing_network', states.RESCUING, 'rescuing',
            'access/native_vlan=333')
        expected_args, expected_kwargs = self._expected_update_call(
            f'Ironic Port {self.port.uuid}', 'access', 333)
        mock_call.assert_called_with(*expected_args, **expected_kwargs)

    def test_remove_rescuing_network_with_config(self):
        mock_call = self._test_remove_network(
            'remove_rescuing_network', states.RESCUING,
            'rescuing', 'access/native_vlan=333')
        mock_call.assert_called_with(
            mock.ANY, '00:11:22:33:44:55', 'GigE1/0/1', 333,
            allowed_vlans=None, default_vlan=None)

    def test_add_inspection_network_without_config(self):
        mock_call = self._test_add_network('add_inspection_network',
                                           states.INSPECTING, 'inspection')
        expected_args, expected_kwargs = self._expected_update_call(
            f'Ironic Port {self.port.uuid}', 'access', 100)
        mock_call.assert_called_with(*expected_args, **expected_kwargs)

    def test_remove_inspection_network_without_config(self):
        mock_call = self._test_remove_network('remove_inspection_network',
                                              states.INSPECTING,
                                              'inspection')
        mock_call.assert_called_with(
            mock.ANY, '00:11:22:33:44:55', 'GigE1/0/1', 100,
            allowed_vlans=None, default_vlan=None)

    def test_add_servicing_network_with_config(self):
        mock_call = self._test_add_network(
            'add_servicing_network', states.SERVICING, 'servicing',
            'access/native_vlan=444')
        expected_args, expected_kwargs = self._expected_update_call(
            f'Ironic Port {self.port.uuid}', 'access', 444)
        mock_call.assert_called_with(*expected_args, **expected_kwargs)

    def test_remove_servicing_network_with_config(self):
        mock_call = self._test_remove_network(
            'remove_servicing_network', states.SERVICING,
            'servicing', 'access/native_vlan=444')
        mock_call.assert_called_with(
            mock.ANY, '00:11:22:33:44:55', 'GigE1/0/1', 444,
            allowed_vlans=None, default_vlan=None)

    def test_configure_tenant_networks(self):
        with mock.patch.object(networking_api, 'update_port',
                               autospec=True) as update_mock:
            update_mock.return_value = {'status': 'updated'}
            with task_manager.acquire(self.context, self.node.uuid) as task:
                task.ports = [self.port]
                self.interface.configure_tenant_networks(task)
        update_mock.assert_called_with(
            self.context, '00:11:22:33:44:55', 'GigE1/0/1',
            f'Ironic Port {self.port.uuid}', 'access', 100,
            allowed_vlans=None, lag_name=None, default_vlan=None)

    def test_unconfigure_tenant_networks(self):
        with mock.patch.object(networking_api, 'reset_port',
                               autospec=True) as reset_mock:
            reset_mock.return_value = {'status': 'reset'}
            with task_manager.acquire(self.context, self.node.uuid) as task:
                task.ports = [self.port]
                self.interface.unconfigure_tenant_networks(task)
        reset_mock.assert_called_with(
            self.context, '00:11:22:33:44:55', 'GigE1/0/1', 100,
            allowed_vlans=None, default_vlan=None)

    def test_validate_rescue_success(self):
        with self._prepare_task(states.RESCUING) as task:
            self.interface.validate_rescue(task)

    def test_validate_inspection_success(self):
        with self._prepare_task(states.INSPECTING) as task:
            self.interface.validate_inspection(task)

    def test_validate_rescue_missing_configuration(self):
        port = self._create_test_port(extra={})
        with task_manager.acquire(self.context, self.node.uuid) as task:
            task.ports = [port]
            self.assertIsNone(
                self.interface._validate_network_requirements(
                    task, 'rescuing'))

    def test_need_power_on(self):
        with task_manager.acquire(self.context, self.node.uuid) as task:
            self.assertFalse(self.interface.need_power_on(task))

    def test_get_node_network_data_present(self):
        self.node.network_data = {'foo': 'bar'}
        self.node.save()
        with task_manager.acquire(self.context, self.node.uuid) as task:
            self.assertEqual({'foo': 'bar'},
                             self.interface.get_node_network_data(task))

    def test_get_node_network_data_default(self):
        """Generates network data from ports when static data not set."""
        self.node.network_data = None
        self.node.save()
        with task_manager.acquire(self.context, self.node.uuid) as task:
            result = self.interface.get_node_network_data(task)
            # Should generate links from the port created in setUp
            self.assertIn('links', result)
            self.assertEqual(1, len(result['links']))
            # Verify the physical link was created
            link = result['links'][0]
            self.assertEqual(self.port.uuid, link['id'])
            self.assertEqual('phy', link['type'])
            self.assertEqual(self.port.address, link['ethernet_mac_address'])
            self.assertEqual(1500, link['mtu'])

    def test_validate_portgroup_members_missing_local_link(self):
        """Portgroup validation raises when member lacks local link."""
        portgroup = self._create_test_portgroup(
            extra={'lag': {
                'mode': 'access',
                'aggregation_mode': 'static'
            }}
        )
        member_port = self._create_test_port(
            portgroup_id=portgroup.id,
            local_link_connection={
                'switch_id': None,
                'port_id': None
            },
            extra={'switchport': {'mode': 'access', 'native_vlan': 100}}
        )
        with task_manager.acquire(self.context, self.node.uuid) as task:
            task.portgroups = [portgroup]
            task.ports = [member_port]
            self.assertRaises(exception.InvalidParameterValue,
                              self.interface.validate, task)


class GetNodeNetworkDataTestCase(IronicNetworkingTestCase):
    """Test cases for get_node_network_data method."""

    def test_get_node_network_data_static_precedence(self):
        """Static network_data takes precedence over generated data."""
        static_data = {
            'links': [{'id': 'static', 'type': 'phy'}],
            'networks': [],
            'services': []
        }
        self.node.network_data = static_data
        self.node.save()

        # Create a port that would generate data if static wasn't present
        self._create_test_port()

        with task_manager.acquire(self.context, self.node.uuid) as task:
            result = self.interface.get_node_network_data(task)

        self.assertEqual(static_data, result)

    def test_get_node_network_data_empty_ports(self):
        """Returns empty links when no ports exist."""
        with task_manager.acquire(self.context, self.node.uuid) as task:
            result = self.interface.get_node_network_data(task)

        self.assertEqual({'links': []}, result)

    def test_get_node_network_data_single_port_access_mode(self):
        """Generates physical link for port in access mode (no VLANs)."""
        mac = self._generate_test_mac()
        port = self._create_test_port(
            address=mac,
            extra={
                'switchport': {
                    'mode': 'access',
                    'native_vlan': 100
                }
            }
        )

        with task_manager.acquire(self.context, self.node.uuid) as task:
            result = self.interface.get_node_network_data(task)

        # Should have only one physical link, no VLANs
        self.assertEqual(1, len(result['links']))
        link = result['links'][0]
        self.assertEqual(port.uuid, link['id'])
        self.assertEqual('phy', link['type'])
        self.assertEqual(mac, link['ethernet_mac_address'])
        self.assertEqual(1500, link['mtu'])
        self.assertNotIn('vif_id', link)

    def test_get_node_network_data_single_port_trunk_mode(self):
        """Generates physical link and VLAN interfaces for trunk mode."""
        mac = self._generate_test_mac()
        allowed_vlans = [100, 200, 300]
        port = self._create_test_port(
            address=mac,
            extra={
                'switchport': {
                    'mode': 'trunk',
                    'native_vlan': 100,
                    'allowed_vlans': allowed_vlans
                }
            }
        )

        with task_manager.acquire(self.context, self.node.uuid) as task:
            result = self.interface.get_node_network_data(task)

        # Should have 1 physical link + 3 VLAN links
        self.assertEqual(4, len(result['links']))

        # Check physical link
        phy_link = result['links'][0]
        self.assertEqual(port.uuid, phy_link['id'])
        self.assertEqual('phy', phy_link['type'])
        self.assertEqual(mac, phy_link['ethernet_mac_address'])
        self.assertEqual(1500, phy_link['mtu'])

        # Check VLAN links
        vlan_links = result['links'][1:]
        self.assertEqual(3, len(vlan_links))
        for i, vlan_link in enumerate(vlan_links):
            expected_vlan_id = allowed_vlans[i]
            self.assertEqual(f'{port.uuid}_vlan_{expected_vlan_id}',
                             vlan_link['id'])
            self.assertEqual('vlan', vlan_link['type'])
            self.assertEqual(mac, vlan_link['vlan_mac_address'])
            self.assertEqual(expected_vlan_id, vlan_link['vlan_id'])
            self.assertEqual(port.uuid, vlan_link['vlan_link'])
            self.assertEqual(1500, vlan_link['mtu'])

    def test_get_node_network_data_single_port_hybrid_mode(self):
        """Generates physical link and VLAN interfaces for hybrid mode."""
        mac = self._generate_test_mac()
        allowed_vlans = [200, 300]
        port = self._create_test_port(
            address=mac,
            extra={
                'switchport': {
                    'mode': 'hybrid',
                    'native_vlan': 100,
                    'allowed_vlans': allowed_vlans
                }
            }
        )

        with task_manager.acquire(self.context, self.node.uuid) as task:
            result = self.interface.get_node_network_data(task)

        # Should have 1 physical link + 2 VLAN links
        self.assertEqual(3, len(result['links']))

        # Check physical link
        phy_link = result['links'][0]
        self.assertEqual(port.uuid, phy_link['id'])
        self.assertEqual('phy', phy_link['type'])

        # Check VLAN links
        vlan_links = result['links'][1:]
        self.assertEqual(2, len(vlan_links))
        vlan_ids = [vl['vlan_id'] for vl in vlan_links]
        self.assertEqual(allowed_vlans, vlan_ids)

    def test_get_node_network_data_multiple_ports(self):
        """Generates links for multiple ports with mixed configurations."""
        mac1 = self._generate_test_mac()
        mac2 = self._generate_test_mac()

        port1 = self._create_test_port(
            address=mac1,
            extra={
                'switchport': {
                    'mode': 'access',
                    'native_vlan': 100
                }
            }
        )

        port2 = self._create_test_port(
            address=mac2,
            extra={
                'switchport': {
                    'mode': 'trunk',
                    'native_vlan': 100,
                    'allowed_vlans': [200, 300]
                }
            }
        )

        with task_manager.acquire(self.context, self.node.uuid) as task:
            result = self.interface.get_node_network_data(task)

        # Port1: 1 phy, Port2: 1 phy + 2 vlans = 4 total
        self.assertEqual(4, len(result['links']))

        # Find links by MAC to verify correct association
        port1_links = [l1 for l1 in result['links']
                       if l1.get('ethernet_mac_address') == mac1
                       or l1.get('vlan_mac_address') == mac1]
        port2_links = [l2 for l2 in result['links']
                       if l2.get('ethernet_mac_address') == mac2
                       or l2.get('vlan_mac_address') == mac2]

        self.assertEqual(1, len(port1_links))  # Only physical
        self.assertEqual(3, len(port2_links))  # Physical + 2 VLANs

        self.assertEqual(port1.uuid, result['links'][0]['id'])
        self.assertEqual(port2.uuid, result['links'][1]['id'])

    def test_get_node_network_data_port_without_mac(self):
        """Skips ports without MAC addresses."""
        port1 = self._create_test_port(
            address=None,
            extra={
                'switchport': {
                    'mode': 'access',
                    'native_vlan': 100
                }
            }
        )

        mac2 = self._generate_test_mac()
        port2 = self._create_test_port(
            address=mac2,
            extra={
                'switchport': {
                    'mode': 'access',
                    'native_vlan': 100
                }
            }
        )

        with task_manager.acquire(self.context, self.node.uuid) as task:
            result = self.interface.get_node_network_data(task)

        # Only port2 should generate a link
        self.assertEqual(1, len(result['links']))
        self.assertEqual(port2.uuid, result['links'][0]['id'])
        self.assertEqual(mac2, result['links'][0]['ethernet_mac_address'])

        # port1 should be ignored
        self.assertNotEqual(port1.uuid, result['links'][0]['id'])

    def test_get_node_network_data_port_without_switchport(self):
        """Generates physical link only for port without switchport config."""
        mac = self._generate_test_mac()
        port = self._create_test_port(address=mac, extra={})

        with task_manager.acquire(self.context, self.node.uuid) as task:
            result = self.interface.get_node_network_data(task)

        # Should have only physical link, no VLANs
        self.assertEqual(1, len(result['links']))
        link = result['links'][0]
        self.assertEqual(port.uuid, link['id'])
        self.assertEqual('phy', link['type'])
        self.assertEqual(mac, link['ethernet_mac_address'])

    def test_get_node_network_data_port_with_empty_allowed_vlans(self):
        """Port with empty allowed_vlans generates only physical link."""
        mac = self._generate_test_mac()
        port = self._create_test_port(
            address=mac,
            extra={
                'switchport': {
                    'mode': 'trunk',
                    'native_vlan': 100,
                    'allowed_vlans': []
                }
            }
        )

        with task_manager.acquire(self.context, self.node.uuid) as task:
            result = self.interface.get_node_network_data(task)

        # Should have only physical link, no VLANs
        self.assertEqual(1, len(result['links']))
        link = result['links'][0]
        self.assertEqual(port.uuid, link['id'])
        self.assertEqual('phy', link['type'])

    def test_get_node_network_data_port_with_none_extra(self):
        """Handles port with None extra field gracefully."""
        mac = self._generate_test_mac()
        port = self._create_test_port(address=mac, extra=None)

        with task_manager.acquire(self.context, self.node.uuid) as task:
            result = self.interface.get_node_network_data(task)

        # Should have only physical link
        self.assertEqual(1, len(result['links']))
        link = result['links'][0]
        self.assertEqual(port.uuid, link['id'])
        self.assertEqual('phy', link['type'])
        self.assertEqual(mac, link['ethernet_mac_address'])



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
"""Unit tests for ``ironic.networking.api``."""
import unittest
from unittest import mock

from ironic.networking import api


class NetworkingApiTestCase(unittest.TestCase):
    """Test cases for helper functions in ``ironic.networking.api``."""

    def setUp(self):
        super().setUp()
        self.addCleanup(setattr, api, "_NETWORKING_API", None)

    def test_get_networking_api_singleton(self):
        with mock.patch(
            "ironic.networking.api.rpcapi.NetworkingAPI",
            autospec=True,
        ) as mock_cls:
            instance = mock_cls.return_value

            result1 = api.get_networking_api()
            result2 = api.get_networking_api()

            self.assertIs(result1, instance)
            self.assertIs(result2, instance)
            mock_cls.assert_called_once_with()

    def test_update_port_delegates_to_rpc(self):
        api_mock = mock.Mock()
        with mock.patch.object(
            api, "get_networking_api", return_value=api_mock,
            autospec=True
        ):
            context = object()
            result = api.update_port(
                context,
                "switch0",
                "eth0",
                "description",
                "access",
                24,
                allowed_vlans=[10],
                lag_name="pc1",
            )

            api_mock.update_port.assert_called_once_with(
                context,
                "switch0",
                "eth0",
                "description",
                "access",
                24,
                allowed_vlans=[10],
                lag_name="pc1",
                default_vlan=None,
            )
            self.assertIs(result, api_mock.update_port.return_value)

    def test_reset_port_delegates_to_rpc(self):
        api_mock = mock.Mock()
        with mock.patch.object(
            api, "get_networking_api", return_value=api_mock,
            autospec=True
        ):
            context = object()
            result = api.reset_port(
                context, "switch1", "eth1", default_vlan=11
            )

            api_mock.reset_port.assert_called_once_with(
                context, "switch1", "eth1", None, allowed_vlans=None,
                default_vlan=11
            )
            self.assertIs(result, api_mock.reset_port.return_value)

    def test_update_lag_delegates_to_rpc(self):
        api_mock = mock.Mock()
        with mock.patch.object(
            api, "get_networking_api", return_value=api_mock,
            autospec=True
        ):
            context = object()
            result = api.update_lag(
                context,
                ["switch1", "switch2"],
                "pc",
                "desc",
                "trunk",
                100,
                "lacp",
                allowed_vlans=[200],
            )

            api_mock.update_lag.assert_called_once_with(
                context,
                ["switch1", "switch2"],
                "pc",
                "desc",
                "trunk",
                100,
                "lacp",
                allowed_vlans=[200],
                default_vlan=None,
            )
            self.assertIs(
                result, api_mock.update_lag.return_value
            )

    def test_delete_lag_delegates_to_rpc(self):
        api_mock = mock.Mock()
        with mock.patch.object(
            api, "get_networking_api", return_value=api_mock,
            autospec=True
        ):
            context = object()
            result = api.delete_lag(
                context, ["switch1"], "pc"
            )

            api_mock.delete_lag.assert_called_once_with(
                context, ["switch1"], "pc"
            )
            self.assertIs(
                result, api_mock.delete_lag.return_value
            )



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
Networking API for other parts of Ironic to use.
"""

from ironic.networking import rpcapi

# Global networking API instance
_NETWORKING_API = None


def get_networking_api():
    """Get the networking API instance.

    :returns: NetworkingAPI instance
    """
    global _NETWORKING_API
    if _NETWORKING_API is None:
        _NETWORKING_API = rpcapi.NetworkingAPI()
    return _NETWORKING_API


def update_port(
    context,
    switch_id,
    port_name,
    description,
    mode,
    native_vlan,
    allowed_vlans=None,
    lag_name=None,
    default_vlan=None
):
    """Update a network switch port configuration.

    This is a convenience function that other parts of Ironic can use
    to update network switch port configurations.

    :param context: request context.
    :param switch_id: Identifier for the switch.
    :param port_name: Name of the port on the switch.
    :param description: Description to set for the port.
    :param mode: Port mode ('access', 'trunk', or 'hybrid').
    :param native_vlan: VLAN ID to be set on the port.
    :param allowed_vlans: List of allowed VLAN IDs to be added(optional).
    :param default_vlan: VLAN ID to removed from the port(optional).
    :param lag_name: LAG name if port is part of a link aggregation group.
    :raises: InvalidParameterValue if validation fails.
    :raises: NetworkError if the network operation fails.
    :returns: Dictionary containing the updated port configuration.
    """
    api = get_networking_api()
    return api.update_port(
        context,
        switch_id,
        port_name,
        description,
        mode,
        native_vlan,
        allowed_vlans=allowed_vlans,
        lag_name=lag_name,
        default_vlan=default_vlan
    )


def reset_port(
    context,
    switch_id,
    port_name,
    native_vlan=None,
    allowed_vlans=None,
    default_vlan=None,
):
    """Reset a network switch port to default configuration.

    This is a convenience function that other parts of Ironic can use
    to reset network switch ports to their default configurations.

    :param context: request context.
    :param switch_id: Identifier for the switch.
    :param port_name: Name of the port on the switch.
    :param native_vlan: VLAN ID to be removed from the port.
    :param allowed_vlans: List of allowed VLAN IDs to be removed(optional).
    :param default_vlan: VLAN ID to restore onto the port(optional).
    :raises: InvalidParameterValue if validation fails.
    :raises: NetworkError if the network operation fails.
    :returns: Dictionary containing the reset port configuration.
    """
    api = get_networking_api()
    return api.reset_port(
        context,
        switch_id,
        port_name,
        native_vlan,
        allowed_vlans=allowed_vlans,
        default_vlan=default_vlan
    )


def update_lag(
    context,
    switch_ids,
    lag_name,
    description,
    mode,
    native_vlan,
    aggregation_mode,
    allowed_vlans=None,
    default_vlan=None
):
    """Update a link aggregation group (LAG) configuration.

    This is a convenience function that other parts of Ironic can use
    to update LAG configurations.

    :param context: request context.
    :param switch_ids: List of switch identifiers.
    :param lag_name: Name of the LAG.
    :param description: Description for the LAG.
    :param mode: LAG mode ('access' or 'trunk').
    :param native_vlan: VLAN ID to be set for the LAG.
    :param aggregation_mode: Aggregation mode (e.g., 'lacp', 'static').
    :param allowed_vlans: List of allowed VLAN IDs to be added (optional).
    :param default_vlan: VLAN ID to removed from the port(optional).
    :raises: InvalidParameterValue if validation fails.
    :raises: NetworkError if the network operation fails.
    :returns: Dictionary containing the updated LAG configuration.
    """
    api = get_networking_api()
    return api.update_lag(
        context,
        switch_ids,
        lag_name,
        description,
        mode,
        native_vlan,
        aggregation_mode,
        allowed_vlans=allowed_vlans,
        default_vlan=default_vlan
    )


def delete_lag(context, switch_ids, lag_name):
    """Delete a link aggregation group (LAG) configuration.

    This is a convenience function that other parts of Ironic can use
    to delete LAG configurations.

    :param context: request context.
    :param switch_ids: List of switch identifiers.
    :param lag_name: Name of the LAG to delete.
    :raises: InvalidParameterValue if validation fails.
    :raises: NetworkError if the network operation fails.
    :returns: Dictionary containing the deletion status.
    """
    api = get_networking_api()
    return api.delete_lag(context, switch_ids, lag_name)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from dataclasses import dataclass
from dataclasses import field

from ironic.common.trait_based_networking.plan import TENANT_VIF_KEY

def default_internal_info(tenant_vif=None):
    return {
        TENANT_VIF_KEY: tenant_vif,
    }

@dataclass
class FauxPortLikeObject(object):
    id: str = "fake_id"
    uuid: str = "fake_uuid"
    address: str = "test"
    category: str = "cat"
    physical_network: str = "test_physnet"
    vendor: str = "fake_vendor"
    internal_info: dict = field(default_factory=default_internal_info)
    # NOTE(clif): Only relevant on ports, this defaults to True in ironic
    available_for_dynamic_portgroup: bool = True
    # NOTE(clif): Only relevant on portgroups
    dynamic_portgroup: bool = False


@dataclass
class FauxNetwork(object):
    id: str = "fake_net_id"
    name: str = "test_network"
    tags: frozenset[str] = field(default_factory=frozenset)

def default_faux_instance_info(traits=None):
    return {
        'traits': traits
    }


@dataclass
class FauxNode(object):
    uuid: str = "fake_node_uuid"
    instance_info: dict = field(
            default_factory=default_faux_instance_info)


@dataclass
class FauxTask(object):
    node: FauxNode
    ports: list[FauxPortLikeObject] = field(default_factory=list)
    portgroups: list[FauxPortLikeObject] = field(default_factory=list)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_config import cfg

from ironic.common.i18n import _

group = cfg.OptGroup(
    name='cifs',
    title='CIFS/SMB Virtual Media Options',
    help=_('Configuration for CIFS/SMB-based virtual media transport. '
           'Operators are responsible for providing a CIFS/SMB server '
           'and ensuring the share specified in base_url is mounted at '
           'share_path on all conductor hosts. The conductor copies '
           'images to share_path, and the BMC accesses them directly '
           'via CIFS/SMB using the base_url. Optional credentials can '
           'be configured and will be passed to the BMC.'))

opts = [
    cfg.StrOpt('base_url',
               help=_('Base URL for CIFS/SMB shares '
                      '(e.g., cifs://server/share or smb://server/share). '
                      'Required for CIFS virtual media support. Must be used '
                      'together with share_path.')),
    cfg.StrOpt('share_path',
               help=_('Local mount point of the CIFS share where images will '
                      'be placed. Required for CIFS virtual media support. '
                      'Must be used together with base_url.')),
    cfg.StrOpt('username',
               help=_('Username for CIFS authentication. If set, credentials '
                      'will be passed to the BMC when inserting virtual '
                      'media.')),
    cfg.StrOpt('password',
               secret=True,
               help=_('Password for CIFS authentication. If set, credentials '
                      'will be passed to the BMC when inserting virtual '
                      'media.')),
]


def register_opts(conf):
    conf.register_group(group)
    conf.register_opts(opts, group='cifs')



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
JSON Schema definitions for the ironic-networking network interface.

This module contains the JSON schemas used to validate switchport and LAG
configuration in port and portgroup 'extra' fields.
"""

SWITCHPORT_SCHEMA = {
    "$schema": "http://json-schema.org/draft-04/schema#",
    "title": "Ironic Networking Network Interface Switchport Configuration "
             "Schema",
    "description": "Schema for validating switchport configuration in port "
                   "'extra' field for ironic networking network interface. "
                   "Switch ID and port name are obtained from "
                   "local_link_connection, and description is generated "
                   "automatically.",
    "type": "object",
    "properties": {
        "mode": {
            "type": "string",
            "description": "Switch port mode configuration",
            "enum": ["access", "trunk", "hybrid"]
        },
        "native_vlan": {
            "type": "integer",
            "description": "Native VLAN ID for the port",
            "minimum": 1,
            "maximum": 4094
        },
        "allowed_vlans": {
            "type": "array",
            "description": "List of allowed VLAN IDs for trunk/hybrid modes",
            "items": {
                "type": "integer",
                "minimum": 1,
                "maximum": 4094
            },
            "uniqueItems": True,
            "maxItems": 100
        },
        "lag_name": {
            "type": "string",
            "description": "Name of the LAG this port belongs to",
            "maxLength": 255
        },
        "mtu": {
            "type": "integer",
            "description": "Ethernet Maximum Transmission Unit of the port",
            "maximum": 9216
        }
    },
    "required": ["mode", "native_vlan"],
    "additionalProperties": False,
    "allOf": [
        {
            "if": {
                "properties": {
                    "mode": {"const": "access"}
                }
            },
            "then": {
                "not": {
                    "required": ["allowed_vlans"]
                }
            }
        },
        {
            "if": {
                "properties": {
                    "mode": {"enum": ["trunk", "hybrid"]}
                }
            },
            "then": {
                "properties": {
                    "allowed_vlans": {
                        "minItems": 1
                    }
                },
                "required": ["allowed_vlans"]
            }
        }
    ]
}

LAG_SCHEMA = {
    "$schema": "http://json-schema.org/draft-04/schema#",
    "title": "Ironic Networking Network Interface LAG Configuration Schema",
    "description": "Schema for validating LAG configuration in portgroup "
                   "'extra' field for ironic networking network interface. "
                   "Different modes have different requirements for VLAN "
                   "configuration.",
    "type": "object",
    "properties": {
        "mode": {
            "type": "string",
            "description": "LAG mode configuration",
            "enum": ["access", "trunk", "hybrid"]
        },
        "native_vlan": {
            "type": "integer",
            "description": "The native VLAN ID for the LAG. If not supplied "
                           "then the switch global default VLAN ID is used.",
            "minimum": 1,
            "maximum": 4094
        },
        "allowed_vlans": {
            "type": "array",
            "description": "List of allowed VLAN IDs for trunk mode",
            "items": {
                "type": "integer",
                "minimum": 1,
                "maximum": 4094
            },
            "uniqueItems": True,
            "maxItems": 100
        },
        "aggregation_mode": {
            "type": "string",
            "description": "Link aggregation protocol mode",
            "enum": ["lacp", "static"]
        },
        "mtu": {
            "type": "integer",
            "description": "Ethernet Maximum Transmission Unit of the port",
            "maximum": 9216
        }
    },
    "required": ["mode", "aggregation_mode"],
    "additionalProperties": False,
    "allOf": [
        {
            "if": {
                "properties": {
                    "mode": {"const": "access"}
                }
            },
            "then": {
                "not": {
                    "required": ["allowed_vlans"]
                }
            }
        },
        {
            "if": {
                "properties": {
                    "mode": {"enum": ["trunk", "hybrid"]}
                }
            },
            "then": {
                "properties": {
                    "allowed_vlans": {
                        "minItems": 1
                    }
                },
                "required": ["allowed_vlans"]
            }
        }
    ]
}



#!/usr/bin/env python3

import os

import json

app_name = os.environ.get("APP", "fake")
app_info = json.loads(os.environ.get("APP_INFO"))
debug = int(os.environ.get("DEBUG", 0))
verify = app_info.get("verify_ca", True)
error = os.environ.get("ERROR", "")

if app_name == "fake":
    # Extensions cannot set file:// URLs so special case the fake driver
    homepage = "file:///drivers/fake/index.html"
else:
    homepage = "file:///drivers/launch/index.html"
    if error:
        homepage += f"?error={error}"

policies = {
    "AppAutoUpdate": False,
    "AutofillAddressEnabled": False,
    "AutofillCreditCardEnabled": False,
    "DisableAppUpdate": True,
    "DisableFirefoxScreenshots": True,
    "DisableFirefoxStudies": True,
    "DisableFirefoxStudies_comment": "Disable Firefox studies",
    "DisablePocket": True,
    "DisableSystemAddonUpdate": True,
    "DisableTelemetry": True,
    "DontCheckDefaultBrowser": True,
    "Homepage": {
        "URL": homepage,
        "StartPage": "homepage",
    },
    "NoDefaultBookmarks": True,
    "OfferToSaveLogins": False,
    "OverrideFirstRunPage": "",
    "OverridePostUpdatePage": "",
    "PasswordManagerEnabled": False,
    "Preferences": {
        "security.ssl.enable_ocsp_stapling": {
            "Value": verify,
            "Status": "locked",
        },
        "dom.disable_open_during_load": {
            "Value": False,
            "Status": "locked",
        },
    },
    "PrintingEnabled": False,
    "PromptForDownloadLocation": False,
    "SanitizeOnShutdown": True,
    "SkipTermsOfUse": True,
    "StartDownloadsInTempDirectory": True,
    "WebsiteFilter": {
        "Block": ["<all_urls>"],
        "Exceptions": ["file:///drivers/*"],
    },
}

if not debug:
    policies.update(
        {
            "BlockAboutConfig": True,
            "BlockAboutAddons": True,
            "BlockAboutProfiles": True,
            "BlockAboutSupport": True,
        }
    )

address = app_info.get("address")
if address:
    policies["WebsiteFilter"]["Exceptions"].append(
        f"{address}/*"
    )

print(json.dumps({"policies": policies}, indent=2))



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from oslo_config import cfg
from oslo_log import log as logging

from ironic.drivers.modules.inspector.hooks import base

LOG = logging.getLogger(__name__)
CONF = cfg.CONF


class CPUSHook(base.InspectionHook):
    """Hook to set node's CPUs count based on the inventory."""

    def __call__(self, task, inventory, plugin_data):
        cpu_count = int(inventory.get('cpu', {}).get('count', 0))
        if cpu_count == 0:
            LOG.warning('No CPU count available for node %s.', task.node.uuid)
            return

        task.node.set_property('cpus', cpu_count)
        task.node.save()



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
Networking service driver factory for loading switch drivers.

This module provides a driver factory system for the networking service,
allowing dynamic loading of switch drivers from external projects via
entry points.
"""

import collections

from oslo_log import log

from ironic.common import driver_factory
from ironic.common import exception
from ironic.common.i18n import _
from ironic.conf import CONF

LOG = log.getLogger(__name__)


class BaseSwitchDriverFactory(driver_factory.BaseDriverFactory):
    """Base factory for discovering, loading and managing switch drivers.

    This factory loads switch drivers from entry points and manages their
    lifecycle. Switch drivers are loaded from external projects and provide
    vendor-specific implementations for network switch management.

    Inherits from common.BaseDriverFactory to ensure consistency with
    Ironic's standard driver factory pattern.
    """

    # Entry point namespace for switch drivers
    _entrypoint_name = "ironic.networking.switch_drivers"

    # Configuration option containing enabled drivers list
    _enabled_driver_list_config_option = "enabled_switch_drivers"

    # Template for logging loaded drivers
    _logging_template = "Loaded the following switch drivers: %s"

    # Use two-phase initialization: load classes first, initialize later
    # This allows config preprocessing before driver initialization
    _invoke_on_load = False

    @classmethod
    def _set_enabled_drivers(cls):
        """Set the list of enabled drivers from configuration.

        Overrides the base implementation to read from the networking
        configuration group instead of the default group.
        """
        enabled_drivers = getattr(
            CONF.ironic_networking, cls._enabled_driver_list_config_option, []
        )

        # Check for duplicated driver entries and warn about them
        counter = collections.Counter(enabled_drivers).items()
        duplicated_drivers = []
        cls._enabled_driver_list = []

        for item, cnt in counter:
            if not item:
                raise exception.ConfigInvalid(
                    error_msg=_(
                        'An empty switch driver was specified in the "%s" '
                        'configuration option. Please fix your ironic.conf '
                        'file.') % cls._enabled_driver_list_config_option
                )
            if cnt > 1:
                duplicated_drivers.append(item)
            cls._enabled_driver_list.append(item)

        if duplicated_drivers:
            raise exception.ConfigInvalid(
                error_msg=_(
                    'The switch driver(s) "%s" is/are duplicated in the '
                    'list of enabled drivers. Please check your '
                    'configuration file.') % ", ".join(duplicated_drivers)
            )

    @classmethod
    def _init_extension_manager(cls):
        """Initialize the extension manager for loading switch drivers.

        Extends the base implementation to handle the case where no
        switch drivers are enabled.
        """
        # Set enabled drivers first
        cls._set_enabled_drivers()

        # Only proceed if we have enabled drivers
        if not cls._enabled_driver_list:
            LOG.info("No switch drivers enabled in configuration")
            return

        # Call parent implementation
        try:
            super()._init_extension_manager()
        except RuntimeError as e:
            if "No suitable drivers found" in str(e):
                LOG.warning(
                    "No switch drivers could be loaded. Check that "
                    "the specified drivers are installed and their "
                    "entry points are correctly defined."
                )
                cls._extension_manager = None
            else:
                raise


def _warn_if_unsupported(ext):
    """Warn if a driver is marked as unsupported."""
    if hasattr(ext.obj, "supported") and not ext.obj.supported:
        LOG.warning(
            'Switch driver "%s" is UNSUPPORTED. It has been '
            "deprecated and may be removed in a future release.",
            ext.name,
        )


class SwitchDriverFactory(BaseSwitchDriverFactory):
    """Factory for loading switch drivers from entry points."""

    pass


# Global factory instance
_switch_driver_factory = None


def get_switch_driver_factory():
    """Get the global switch driver factory instance."""
    global _switch_driver_factory
    if _switch_driver_factory is None:
        _switch_driver_factory = SwitchDriverFactory()
    return _switch_driver_factory


def get_switch_driver(driver_name):
    """Get a switch driver instance by name.

    :param driver_name: Name of the switch driver to retrieve.
    :returns: Instance of the switch driver.
    :raises: DriverNotFound if the driver is not found.
    """
    factory = get_switch_driver_factory()
    return factory.get_driver(driver_name)


def list_switch_drivers():
    """Get a list of all available switch driver names.

    :returns: List of switch driver names.
    """
    factory = get_switch_driver_factory()
    return factory.names


def switch_drivers():
    """Get all switch drivers as a dictionary.

    :returns: Dictionary mapping driver name to driver instance.
    """
    factory = get_switch_driver_factory()
    return dict(factory.items())



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

from oslo_config import cfg

from ironic.command import conductor
from ironic.tests.unit.db import base as db_base


class ConductorStartTestCase(db_base.DbTestCase):

    @mock.patch.object(conductor, 'LOG', autospec=True)
    def test_warn_about_unsafe_shred_parameters_defaults(self, log_mock):
        conductor.warn_about_unsafe_shred_parameters(cfg.CONF)
        self.assertFalse(log_mock.warning.called)

    @mock.patch.object(conductor, 'LOG', autospec=True)
    def test_warn_about_unsafe_shred_parameters_zeros(self, log_mock):
        cfg.CONF.set_override('shred_random_overwrite_iterations', 0, 'deploy')
        cfg.CONF.set_override('shred_final_overwrite_with_zeros', True,
                              'deploy')
        conductor.warn_about_unsafe_shred_parameters(cfg.CONF)
        self.assertFalse(log_mock.warning.called)

    @mock.patch.object(conductor, 'LOG', autospec=True)
    def test_warn_about_unsafe_shred_parameters_random_no_zeros(self,
                                                                log_mock):
        cfg.CONF.set_override('shred_random_overwrite_iterations', 1, 'deploy')
        cfg.CONF.set_override('shred_final_overwrite_with_zeros', False,
                              'deploy')
        conductor.warn_about_unsafe_shred_parameters(cfg.CONF)
        self.assertFalse(log_mock.warning.called)

    @mock.patch.object(conductor, 'LOG', autospec=True)
    def test_warn_about_unsafe_shred_parameters_produces_a_warning(self,
                                                                   log_mock):
        cfg.CONF.set_override('shred_random_overwrite_iterations', 0, 'deploy')
        cfg.CONF.set_override('shred_final_overwrite_with_zeros', False,
                              'deploy')
        conductor.warn_about_unsafe_shred_parameters(cfg.CONF)
        self.assertTrue(log_mock.warning.called)

    @mock.patch.object(conductor, 'LOG', autospec=True)
    def test_warn_about_inconsistent_kernel_ramdisk_no_config(self, log_mock):
        # Test when all config dicts are empty (default state)
        conductor.warn_about_inconsistent_kernel_ramdisk(cfg.CONF)
        self.assertFalse(log_mock.warning.called)

    @mock.patch.object(conductor, 'LOG', autospec=True)
    def test_warn_about_inconsistent_kernel_ramdisk_consistent(self, log_mock):
        # Test when kernel and ramdisk configs have matching architectures
        cfg.CONF.set_override('deploy_kernel_by_arch',
                              {'x86_64': 'kernel1', 'aarch64': 'kernel2'},
                              'conductor')
        cfg.CONF.set_override('deploy_ramdisk_by_arch',
                              {'x86_64': 'ramdisk1', 'aarch64': 'ramdisk2'},
                              'conductor')
        cfg.CONF.set_override('rescue_kernel_by_arch',
                              {'x86_64': 'rkernel1'},
                              'conductor')
        cfg.CONF.set_override('rescue_ramdisk_by_arch',
                              {'x86_64': 'rramdisk1'},
                              'conductor')
        conductor.warn_about_inconsistent_kernel_ramdisk(cfg.CONF)
        self.assertFalse(log_mock.warning.called)

    @mock.patch.object(conductor, 'LOG', autospec=True)
    def test_warn_about_inconsistent_kernel_ramdisk_kernel_only(self,
                                                                log_mock):
        # Test when kernel has architectures that ramdisk doesn't
        cfg.CONF.set_override('deploy_kernel_by_arch',
                              {'x86_64': 'kernel1', 'aarch64': 'kernel2'},
                              'conductor')
        cfg.CONF.set_override('deploy_ramdisk_by_arch',
                              {'x86_64': 'ramdisk1'},
                              'conductor')
        conductor.warn_about_inconsistent_kernel_ramdisk(cfg.CONF)
        self.assertEqual(1, log_mock.warning.call_count)
        warning_call = log_mock.warning.call_args[0]
        # Check the warning message format and arguments
        self.assertIn('[conductor]%s', warning_call[0])
        self.assertEqual('deploy_kernel_by_arch', warning_call[1])
        self.assertEqual('aarch64', warning_call[2])
        self.assertEqual('deploy_ramdisk_by_arch', warning_call[3])
        self.assertEqual('provisioning', warning_call[4])

    @mock.patch.object(conductor, 'LOG', autospec=True)
    def test_warn_about_inconsistent_kernel_ramdisk_ramdisk_only(self,
                                                                 log_mock):
        # Test when ramdisk has architectures that kernel doesn't
        cfg.CONF.set_override('rescue_kernel_by_arch',
                              {'x86_64': 'kernel1'},
                              'conductor')
        cfg.CONF.set_override('rescue_ramdisk_by_arch',
                              {'x86_64': 'ramdisk1', 'ppc64le': 'ramdisk2'},
                              'conductor')
        conductor.warn_about_inconsistent_kernel_ramdisk(cfg.CONF)
        self.assertEqual(1, log_mock.warning.call_count)
        warning_call = log_mock.warning.call_args[0]
        # Check the warning message format and arguments
        self.assertIn('[conductor]%s', warning_call[0])
        self.assertEqual('rescue_ramdisk_by_arch', warning_call[1])
        self.assertEqual('ppc64le', warning_call[2])
        self.assertEqual('rescue_kernel_by_arch', warning_call[3])
        self.assertEqual('rescue', warning_call[4])

    @mock.patch.object(conductor, 'LOG', autospec=True)
    def test_warn_about_inconsistent_kernel_ramdisk_both_bad(self,
                                                             log_mock):
        # Test when both kernel and ramdisk have mismatched architectures
        cfg.CONF.set_override('deploy_kernel_by_arch',
                              {'x86_64': 'kernel1', 'aarch64': 'kernel2'},
                              'conductor')
        cfg.CONF.set_override('deploy_ramdisk_by_arch',
                              {'x86_64': 'ramdisk1', 'ppc64le': 'ramdisk2'},
                              'conductor')
        conductor.warn_about_inconsistent_kernel_ramdisk(cfg.CONF)
        self.assertEqual(2, log_mock.warning.call_count)
        # Check that both warnings were issued with correct parameters
        warning_args = [call[0] for call in log_mock.warning.call_args_list]
        # First warning about kernel_only (aarch64)
        self.assertEqual('deploy_kernel_by_arch', warning_args[0][1])
        self.assertEqual('aarch64', warning_args[0][2])
        # Second warning about ramdisk_only (ppc64le)
        self.assertEqual('deploy_ramdisk_by_arch', warning_args[1][1])
        self.assertEqual('ppc64le', warning_args[1][2])

    @mock.patch.object(conductor, 'LOG', autospec=True)
    def test_warn_about_inconsistent_kernel_ramdisk_multiple_configs(self,
                                                                     log_mock):
        # Test multiple config pairs with inconsistencies
        cfg.CONF.set_override('deploy_kernel_by_arch',
                              {'x86_64': 'kernel1', 'aarch64': 'kernel2'},
                              'conductor')
        cfg.CONF.set_override('deploy_ramdisk_by_arch',
                              {'x86_64': 'ramdisk1'},
                              'conductor')
        cfg.CONF.set_override('rescue_kernel_by_arch',
                              {'ppc64le': 'rkernel1'},
                              'conductor')
        cfg.CONF.set_override('rescue_ramdisk_by_arch',
                              {'ppc64le': 'rramdisk1', 's390x': 'rramdisk2'},
                              'conductor')
        conductor.warn_about_inconsistent_kernel_ramdisk(cfg.CONF)
        self.assertEqual(2, log_mock.warning.call_count)
        # Verify both deploy and rescue warnings were issued
        warning_args = [call[0] for call in log_mock.warning.call_args_list]
        # Check operation names in warnings
        operations = [args[4] for args in warning_args]
        self.assertIn('provisioning', operations)
        self.assertIn('rescue', operations)



# Copyright 2018 DMTF. All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_log import log
import sushy

from ironic.common import exception
from ironic.common.i18n import _
from ironic.common import metrics_utils
from ironic.common import states
from ironic.conductor import periodics
from ironic.conductor import utils as manager_utils
from ironic.conf import CONF
from ironic.drivers import base
from ironic.drivers.modules import deploy_utils
from ironic.drivers.modules.redfish import utils as redfish_utils
from ironic import objects

LOG = log.getLogger(__name__)

METRICS = metrics_utils.get_metrics_logger(__name__)

registry_fields = ('attribute_type', 'allowable_values', 'lower_bound',
                   'max_length', 'min_length', 'read_only',
                   'reset_required', 'unique', 'upper_bound')

BIOS_REBOOT_STATES = {
    sushy.BootProgressStates.OS_BOOT_STARTED,
    sushy.BootProgressStates.OS_RUNNING,
}
_DII_STATE = 'redfish_bios_state'
_REBOOT_REQUESTED = 'reboot_requested'
_REQUESTED_BIOS_ATTRS = 'requested_bios_attrs'


class RedfishBIOS(base.BIOSInterface):

    _APPLY_CONFIGURATION_ARGSINFO = {
        'settings': {
            'description': (
                'A list of BIOS settings to be applied'
            ),
            'required': True
        }
    }

    def _parse_allowable_values(self, node, allowable_values):
        """Convert the BIOS registry allowable_value list to expected strings

        :param allowable_values: list of dicts of valid values for enumeration
        :returns: list containing only allowable value names
        """

        # Get name from ValueName if it exists, otherwise use ValueDisplayName
        new_list = []
        for dic in allowable_values:
            key = dic.get('ValueName') or dic.get('ValueDisplayName')
            if key:
                new_list.append(key)
            else:
                LOG.warning('Cannot detect the value name for enumeration '
                            'item %(item)s for node %(node)s',
                            {'item': dic, 'node': node.uuid})

        return new_list

    def cache_bios_settings(self, task):
        """Store or update the current BIOS settings for the node.

        Get the current BIOS settings and store them in the bios_settings
        database table.

        :param task: a TaskManager instance containing the node to act on.
        :raises: RedfishConnectionError when it fails to connect to Redfish
        :raises: RedfishError on an error from the Sushy library
        :raises: UnsupportedDriverExtension if the system does not support BIOS
            settings
        """

        node_id = task.node.id
        system = redfish_utils.get_system(task.node)
        try:
            attributes = system.bios.attributes
        except sushy.exceptions.MissingAttributeError:
            error_msg = _('Cannot fetch BIOS attributes for node %s, '
                          'BIOS settings are not supported.') % task.node.uuid
            LOG.error(error_msg)
            raise exception.UnsupportedDriverExtension(error_msg)

        settings = []
        # Convert Redfish BIOS attributes to Ironic BIOS settings
        if attributes:
            settings = [{'name': k, 'value': v}
                        for k, v in attributes.items()]

        # Get the BIOS Registry
        registry_attributes = []
        try:
            bios_registry = system.bios.get_attribute_registry()

            if bios_registry:
                registry_attributes = bios_registry.registry_entries.attributes

        except Exception as e:
            LOG.info('Cannot get BIOS Registry attributes for node %(node)s, '
                     'Error %(exc)s.', {'node': task.node.uuid, 'exc': e})

        # TODO(bfournier): use a common list for registry field names
        # e.g. registry_fields = objects.BIOSSetting.registry_fields

        # The BIOS registry will contain more entries than the BIOS settings
        # Find the registry entry matching the setting name and get the fields
        if registry_attributes:
            for setting in settings:
                reg = next((r for r in registry_attributes
                            if r.name == setting['name']), None)
                fields = [attr for attr in dir(reg)
                          if not attr.startswith("_")]
                settable_keys = [f for f in fields if f in registry_fields]
                # Set registry fields to current values
                for k in settable_keys:
                    setting[k] = getattr(reg, k, None)
                    if k == "allowable_values" and isinstance(setting[k],
                                                              list):
                        setting[k] = self._parse_allowable_values(
                            task.node, setting[k])

        LOG.debug('Cache BIOS settings for node %(node_uuid)s',
                  {'node_uuid': task.node.uuid})

        create_list, update_list, delete_list, nochange_list = (
            objects.BIOSSettingList.sync_node_setting(
                task.context, node_id, settings))

        if create_list:
            objects.BIOSSettingList.create(
                task.context, node_id, create_list)
        if update_list:
            objects.BIOSSettingList.save(
                task.context, node_id, update_list)
        if delete_list:
            delete_names = [d['name'] for d in delete_list]
            objects.BIOSSettingList.delete(
                task.context, node_id, delete_names)

    @base.service_step(priority=0, requires_ramdisk=False)
    @base.clean_step(priority=0, requires_ramdisk=False)
    @base.deploy_step(priority=0)
    @base.cache_bios_settings
    def factory_reset(self, task):
        """Reset the BIOS settings of the node to the factory default.

        :param task: a TaskManager instance containing the node to act on.
        :raises: RedfishConnectionError when it fails to connect to Redfish
        :raises: RedfishError on an error from the Sushy library
        """
        system = redfish_utils.get_system(task.node)
        try:
            bios = system.bios
        except sushy.exceptions.MissingAttributeError:
            error_msg = (_('Redfish BIOS factory reset failed for node '
                           '%s, because BIOS settings are not supported.') %
                         task.node.uuid)
            LOG.error(error_msg)
            raise exception.RedfishError(error=error_msg)

        node = task.node
        info = node.driver_internal_info
        bios_state = info.get(_DII_STATE) or {}
        reboot_requested = bios_state.get(_REBOOT_REQUESTED, False)
        if not reboot_requested:
            LOG.debug('Factory reset BIOS configuration for node %(node)s',
                      {'node': node.uuid})
            try:
                bios.reset_bios()
            except sushy.exceptions.SushyError as e:
                error_msg = (_('Redfish BIOS factory reset failed for node '
                               '%(node)s. Error: %(error)s') %
                             {'node': node.uuid, 'error': e})
                LOG.error(error_msg)
                raise exception.RedfishError(error=error_msg)

            self._set_reboot_requested(
                task,
                attributes=None)
            return self.post_reset(task)
        else:
            current_attrs = bios.attributes
            LOG.debug('Post factory reset, BIOS configuration for node '
                      '%(node_uuid)s: %(attrs)r',
                      {'node_uuid': node.uuid, 'attrs': current_attrs})
            self._clear_reboot_requested(task)

    @base.service_step(priority=0, argsinfo=_APPLY_CONFIGURATION_ARGSINFO,
                       requires_ramdisk=False)
    @base.clean_step(priority=0, argsinfo=_APPLY_CONFIGURATION_ARGSINFO,
                     requires_ramdisk=False)
    @base.deploy_step(priority=0, argsinfo=_APPLY_CONFIGURATION_ARGSINFO)
    @base.cache_bios_settings
    def apply_configuration(self, task, settings):
        """Apply the BIOS settings to the node.

        :param task: a TaskManager instance containing the node to act on.
        :param settings: a list of BIOS settings to be updated.
        :raises: RedfishConnectionError when it fails to connect to Redfish
        :raises: RedfishError on an error from the Sushy library
        """

        system = redfish_utils.get_system(task.node)
        try:
            bios = system.bios
        except sushy.exceptions.MissingAttributeError:
            error_msg = (_('Redfish BIOS factory reset failed for node '
                           '%s, because BIOS settings are not supported.') %
                         task.node.uuid)
            LOG.error(error_msg)
            raise exception.RedfishError(error=error_msg)

        # Convert Ironic BIOS settings to Redfish BIOS attributes
        attributes = {s['name']: s['value'] for s in settings}

        info = task.node.driver_internal_info
        bios_state = info.get(_DII_STATE) or {}
        reboot_requested = bios_state.get(_REBOOT_REQUESTED, False)

        if not reboot_requested:
            # Check if all requested settings already match the current
            # BIOS values.  When a client of the Ironic API re-sends
            # the same request after a successful apply, this avoids an
            # unnecessary reboot cycle.
            current_attrs = bios.attributes or {}
            try:
                pending_attrs = bios.pending_attributes
            except (sushy.exceptions.SushyError, AttributeError):
                pending_attrs = {}
            if not isinstance(pending_attrs, dict):
                pending_attrs = {}
            all_match = True
            for s in settings:
                name, value = s['name'], s['value']
                if current_attrs.get(name) != value:
                    all_match = False
                    break
                if name in pending_attrs:
                    if pending_attrs[name] != value:
                        # A conflicting pending change exists.
                        all_match = False
                        break
            if all_match:
                LOG.info('All requested BIOS settings for node '
                         '%(node_uuid)s already match the current '
                         'values, skipping apply and reboot.',
                         {'node_uuid': task.node.uuid})
                return

            # Step 1: Apply settings and issue a reboot
            LOG.debug('Apply BIOS configuration for node %(node_uuid)s: '
                      '%(settings)r', {'node_uuid': task.node.uuid,
                                       'settings': settings})
            apply_time = None
            try:
                if bios.supported_apply_times and (
                        sushy.APPLY_TIME_ON_RESET in
                        bios.supported_apply_times):
                    apply_time = sushy.APPLY_TIME_ON_RESET
            except AttributeError:
                LOG.warning('SupportedApplyTimes attribute missing for BIOS'
                            ' configuration on node %(node_uuid)s: ',
                            {'node_uuid': task.node.uuid})

            try:
                bios.set_attributes(attributes, apply_time=apply_time)
            except sushy.exceptions.SushyError as e:
                error_msg = (_('Redfish BIOS apply configuration failed for '
                               'node %(node)s. Error: %(error)s') %
                             {'node': task.node.uuid, 'error': e})
                LOG.error(error_msg)
                raise exception.RedfishError(error=error_msg)

            self._set_reboot_requested(
                task,
                attributes)
            return self.post_configuration(task, settings)
        else:
            # Step 2: Verify requested BIOS settings applied
            requested_attrs = (
                bios_state.get(_REQUESTED_BIOS_ATTRS)
                or info.get(_REQUESTED_BIOS_ATTRS))
            LOG.debug('Verify BIOS configuration for node %(node_uuid)s: '
                      '%(attrs)r', {'node_uuid': task.node.uuid,
                                    'attrs': requested_attrs})
            self._clear_reboot_requested(task)
            attrs_not_updated = self._get_unapplied_bios_attrs(
                task, requested_attrs, bios)
            if attrs_not_updated:
                LOG.debug('BIOS settings %(attrs)s for node %(node_uuid)s '
                          'not updated.', {'attrs': attrs_not_updated,
                                           'node_uuid': task.node.uuid})
                self._set_step_failed(task, attrs_not_updated)
            else:
                LOG.debug('Verification of BIOS settings for node '
                          '%(node_uuid)s successful.',
                          {'node_uuid': task.node.uuid})

    def post_reset(self, task):
        """Perform post reset action to apply the BIOS factory reset.

        Extension point to allow vendor implementations to extend this class
        and override this method to perform a custom action to apply the BIOS
        factory reset to the Redfish service. The default implementation
        performs a reboot.

        :param task: a TaskManager instance containing the node to act on.
        """
        return deploy_utils.reboot_to_finish_step(task)

    def post_configuration(self, task, settings):
        """Perform post configuration action to store the BIOS settings.

        Extension point to allow vendor implementations to extend this class
        and override this method to perform a custom action to write the BIOS
        settings to the Redfish service. The default implementation performs
        a reboot.

        :param task: a TaskManager instance containing the node to act on.
        :param settings: a list of BIOS settings to be updated.
        """
        return deploy_utils.reboot_to_finish_step(task)

    def get_properties(self):
        """Return the properties of the interface.

        :returns: dictionary of <property name>:<property description> entries.
        """
        return redfish_utils.COMMON_PROPERTIES.copy()

    def validate(self, task):
        """Validates the driver information needed by the redfish driver.

        :param task: a TaskManager instance containing the node to act on.
        :raises: InvalidParameterValue on malformed parameter(s)
        :raises: MissingParameterValue on missing parameter(s)
        """
        redfish_utils.parse_driver_info(task.node)

    def _get_unapplied_bios_attrs(self, task, requested_attrs, bios):
        """Return requested BIOS attrs that have not yet been applied.

        :param task: a TaskManager instance containing the node to act on.
        :param requested_attrs: the requested BIOS attributes to update.
        :param bios: BIOS resource object from the system.
        :returns: dict of attributes not yet applied, empty if all applied.
        """
        bios.refresh(force=True)
        current_attrs = bios.attributes
        pending_attrs = bios.pending_attributes
        if not isinstance(pending_attrs, dict):
            pending_attrs = {}
        attrs_not_updated = {}
        for attr in requested_attrs:
            if requested_attrs[attr] != current_attrs.get(attr):
                attrs_not_updated[attr] = requested_attrs[attr]
            elif attr in pending_attrs:
                attrs_not_updated[attr] = requested_attrs[attr]
        return attrs_not_updated

    def _set_reboot_requested(self, task, attributes):
        """Set driver_internal_info flags for reboot requested.

        :param task: a TaskManager instance containing the node to act on.
        :param attributes: the requested BIOS attributes to update.
        """
        node = task.node
        bios_state = {_REBOOT_REQUESTED: True}
        if attributes:
            bios_state[_REQUESTED_BIOS_ATTRS] = attributes
        node.set_driver_internal_info(_DII_STATE, bios_state)
        node.save()
        disable_ramdisk = deploy_utils.is_ramdisk_disabled(node)
        # polling=True tells the IPA heartbeat handler to stand down so
        # that only our periodic task (_query_bios_apply_status) drives
        # step completion.  When ramdisk is active (polling=False), both
        # the heartbeat and the periodic task may race; the exclusive
        # lock in continue_node_clean/deploy/service ensures only one
        # wins.
        deploy_utils.set_async_step_flags(task.node, reboot=True,
                                          skip_current_step=False,
                                          polling=disable_ramdisk)

    def _clear_reboot_requested(self, task):
        """Clear driver_internal_info flags after reboot completed.

        :param task: a TaskManager instance containing the node to act on.
        """
        node = task.node
        node.del_driver_internal_info(_DII_STATE)
        # Drop legacy fields if present from older runs.
        node.del_driver_internal_info('post_bios_reboot_requested')
        node.del_driver_internal_info(_REQUESTED_BIOS_ATTRS)
        node.save()

    def _set_step_failed(self, task, attrs_not_updated):
        """Fail the cleaning or deployment step and log the error.

        :param task: a TaskManager instance containing the node to act on.
        :param attrs_not_updated: the BIOS attributes that were not updated.
        """
        error_msg = (_('Redfish BIOS apply_configuration step failed for node '
                       '%(node)s. Attributes %(attrs)s are not updated.') %
                     {'node': task.node.uuid, 'attrs': attrs_not_updated})
        last_error = (_('Redfish BIOS apply_configuration step failed. '
                        'Attributes %(attrs)s are not updated.') %
                      {'attrs': attrs_not_updated})
        deploy_utils.step_error_handler(task, error_msg, last_error)

    @METRICS.timer('RedfishBIOS._query_bios_apply_status')
    @periodics.node_periodic(
        purpose='checking async redfish BIOS apply/reset status',
        spacing=CONF.redfish.firmware_update_status_interval,
        filters={'reserved': False,
                 'provision_state_in': {states.CLEANWAIT,
                                        states.SERVICEWAIT,
                                        states.DEPLOYWAIT}},
        predicate_extra_fields=['driver_internal_info'],
        predicate=lambda n: n.driver_internal_info.get(_DII_STATE),
    )
    def _query_bios_apply_status(self, task, manager, context):
        self._check_node_redfish_bios_apply(task)

    @METRICS.timer('RedfishBIOS._check_node_redfish_bios_apply')
    def _check_node_redfish_bios_apply(self, task):
        node = task.node

        bios_state = node.driver_internal_info.get(_DII_STATE) or {}
        if not bios_state:
            LOG.debug('BIOS state cleared for node %(node)s before periodic '
                      'could process it (likely a timeout race).',
                      {'node': node.uuid})
            return

        try:
            system = redfish_utils.get_system(node)
        except (exception.RedfishError,
                exception.RedfishConnectionError,
                sushy.exceptions.SushyError) as e:
            LOG.warning('Unable to query Redfish system for node %(node)s '
                        'while waiting for BIOS reboot completion: '
                        '%(error)s',
                        {'node': node.uuid, 'error': e})
            return

        last_state = system.boot_progress.last_state
        if last_state is not None:
            # BootProgress is reported — touch provisioning to prevent the
            # global timeout from firing while we observe meaningful progress.
            node.touch_provisioning()
            if last_state not in BIOS_REBOOT_STATES:
                LOG.debug('Node %(node)s boot progress: %(state)s. '
                          'Waiting for boot progress to reach OS started.',
                          {'node': node.uuid, 'state': last_state})
                return
        # When BootProgress is unavailable (last_state is None), fall
        # through to the attrs check below. Do NOT touch provisioning
        # so the global timeout remains the safety net.

        requested_attrs = bios_state.get(_REQUESTED_BIOS_ATTRS)
        if requested_attrs:
            attrs_not_updated = self._get_unapplied_bios_attrs(
                task, requested_attrs, system.bios)
            if attrs_not_updated:
                LOG.debug('BIOS settings %(attrs)s for node %(node_uuid)s '
                          'not yet applied; continue polling.',
                          {'attrs': attrs_not_updated,
                           'node_uuid': node.uuid})
                return

        LOG.info('Detected post-BIOS reboot completion for node %(node)s, '
                 'resuming the current step.',
                 {'node': node.uuid})

        if node.clean_step:
            manager_utils.notify_conductor_resume_clean(task)
        elif node.service_step:
            manager_utils.notify_conductor_resume_service(task)
        elif node.deploy_step:
            manager_utils.notify_conductor_resume_deploy(task)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from ironic.common.i18n import _
from ironic.common.trait_based_networking.config_file import ConfigFile
from ironic.conf import CONF

from oslo_log import log

import os
import threading


LOG = log.getLogger(__name__)

_LOADER_LOCK = threading.Lock()
_CACHED_CONFIG_LOADER = None


def tbn_config_file_traits():
    """Get TBN traits from the configured YAML file.

    Thread-safe way to retrieve the configured and parsed TBN traits.
    """
    global _CACHED_CONFIG_LOADER
    with _LOADER_LOCK:
        if _CACHED_CONFIG_LOADER is None:
            _CACHED_CONFIG_LOADER = ConfigLoader()
        _CACHED_CONFIG_LOADER.refresh()
        return _CACHED_CONFIG_LOADER.traits


def is_config_valid():
    with _LOADER_LOCK:
        if _CACHED_CONFIG_LOADER is None:
            return False
        return _CACHED_CONFIG_LOADER.valid


class ConfigLoader(object):
    """Extends ConfigFile and provides automatic config file change detection

    Not intended for consumption outside of this module. Use the module level
    tbn_config_file_traits() function to retrieve configured and parsed traits.
    """
    def __init__(self):
        self._last_mtime = None
        self._config_file = None
        self._valid = False
        self.refresh()

    def _has_file_changed(self):
        """Returns True if the underlying config file's mtime has changed."""
        new_mtime = os.path.getmtime(
            CONF.conductor.trait_based_networking_config_file)

        if self._last_mtime is None:
            self._last_mtime = new_mtime
            return True

        if self._last_mtime != new_mtime:
            self._last_mtime = new_mtime
            return True

        return False

    def refresh(self):
        if self._has_file_changed():
            self._load_tbn_config()

    @property
    def valid(self):
        return self._valid

    @property
    def traits(self):
        if self._config_file is None:
            return []
        return self._config_file.traits()

    def _load_tbn_config(self):
        """Load Trait Based Networking configuration file for later use"""
        if not CONF.conductor.enable_trait_based_networking:
            LOG.debug('Trait Based Networking not enabled, skipping loading '
                      'configuration file.')
            return

        LOG.info(('Loading Trait Based Networking configuration located at '
                  '%(file_location)s'),
                 {'file_location':
                  CONF.conductor.trait_based_networking_config_file})
        try:
            self._config_file = ConfigFile(
                CONF.conductor.trait_based_networking_config_file)
        except OSError as err:
            self._valid = False
            LOG.error(('Failed to load Trait Based Networking configuration '
                       'file located at \'%(file)s\', error: %(err)s'), {
                           'file':
                               CONF.conductor.
                               trait_based_networking_config_file,
                           'err': err})
            return

        valid, reasons = self._config_file.validate()

        if not valid:
            self._valid = False
            LOG.error(('Configuration file for Trait Based Networking '
                       'located at \'%(file)s\' is invalid. Reasons '
                       'follow:'), {
                        'file': \
                        CONF.conductor.trait_based_networking_config_file
                     })
            for reason in reasons:
                LOG.error(reason)

            return

        # NOTE(clif) If we made it here the configuration should parse.
        self._config_file.parse()

        self._valid = True
        LOG.info(_('Successfully loaded and parsed configuration for '
                   'Trait Based Networking.'))



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Unit tests for networking RPC API."""

from oslo_config import cfg
import unittest.mock as mock

from ironic.common import exception
from ironic.common import rpc
from ironic.networking import rpcapi
from ironic.tests import base as test_base

CONF = cfg.CONF


class TestNetworkingAPI(test_base.TestCase):
    """Test cases for NetworkingAPI RPC client."""

    def setUp(self):
        super(TestNetworkingAPI, self).setUp()
        self.context = mock.Mock()
        self.api = rpcapi.NetworkingAPI()

    def test_init_default_topic(self):
        """Test NetworkingAPI initialization with default topic."""
        with mock.patch.object(rpc, "NETWORKING_TOPIC", "test-topic"):
            api = rpcapi.NetworkingAPI()
            self.assertEqual("test-topic", api.topic)

    def test_init_custom_topic(self):
        """Test NetworkingAPI initialization with custom topic."""
        api = rpcapi.NetworkingAPI(topic="custom-topic")
        self.assertEqual("custom-topic", api.topic)

    def test_get_topic(self):
        """Test get_topic method."""
        with mock.patch.object(rpc, "NETWORKING_TOPIC", "test-topic"):
            api = rpcapi.NetworkingAPI()
            self.assertEqual("test-topic", api.get_topic())

    @mock.patch.object(rpc, "NETWORKING_TOPIC", "test-topic")
    def test_init_with_none_transport(self):
        """Test initialization with rpc_transport=none."""
        CONF.set_override("rpc_transport", "none")
        api = rpcapi.NetworkingAPI()
        self.assertIsNone(api.client)

    @mock.patch.object(rpc, "NETWORKING_TOPIC", "test-topic")
    def test_init_with_json_rpc_transport(self):
        """Test initialization with json-rpc transport."""
        CONF.set_override("rpc_transport", "json-rpc")
        api = rpcapi.NetworkingAPI()
        self.assertIsNotNone(api.client)
        # Ensure the client is configured to use networking group
        self.assertEqual("ironic_networking_json_rpc", api.client.conf_group)

    @mock.patch.object(rpc, "NETWORKING_TOPIC", "test-topic")
    def test_init_with_oslo_messaging_transport(self):
        """Test initialization with oslo.messaging transport."""
        CONF.set_override("rpc_transport", "oslo")
        api = rpcapi.NetworkingAPI()
        self.assertIsNotNone(api.client)

    @mock.patch.object(rpc, "NETWORKING_TOPIC", "test-topic")
    def test_init_with_json_rpc_uses_networking_host(self):
        """Test initialization with json-rpc uses networking host for topic."""
        CONF.set_override("rpc_transport", "json-rpc")
        CONF.set_override(
            "host_ip", "test-networking-host",
            group="ironic_networking_json_rpc"
        )
        CONF.set_override("port", 8089, group="ironic_networking_json_rpc")
        api = rpcapi.NetworkingAPI()
        self.assertEqual("ironic.test-networking-host:8089", api.topic)
        self.assertIsNotNone(api.client)

    @mock.patch.object(rpc, "NETWORKING_TOPIC", "test-topic")
    def test_init_with_oslo_messaging_uses_default_topic(self):
        """Test initialization with oslo.messaging uses default topic."""
        CONF.set_override("rpc_transport", "oslo")
        CONF.set_override("host", "test-networking-host")
        api = rpcapi.NetworkingAPI()
        self.assertEqual("test-topic", api.topic)
        self.assertIsNotNone(api.client)

    @mock.patch.object(rpc, "NETWORKING_TOPIC", "test-topic")
    def test_init_with_custom_topic_overrides_networking_host(self):
        """Test custom topic overrides networking host even for JSON-RPC."""
        CONF.set_override("rpc_transport", "json-rpc")
        CONF.set_override("host", "test-networking-host")
        api = rpcapi.NetworkingAPI(topic="custom-topic")
        self.assertEqual("custom-topic", api.topic)
        self.assertIsNotNone(api.client)

    def test_prepare_call_none_client(self):
        """Test _prepare_call with None client raises exception."""
        self.api.client = None
        exc = self.assertRaises(
            exception.ServiceUnavailable, self.api._prepare_call, topic="test"
        )
        self.assertIn("Cannot use 'none' RPC", str(exc))

    @mock.patch.object(rpcapi.NetworkingAPI, "_prepare_call", autospec=True)
    def test_update_port_success(self, mock_prepare):
        """Test successful update_port call."""
        mock_cctxt = mock.Mock()
        mock_prepare.return_value = mock_cctxt
        mock_cctxt.call.return_value = {"status": "configured"}

        result = self.api.update_port(
            self.context,
            "switch-01",
            "port-01",
            "Test port",
            "access",
            100,
            allowed_vlans=None,
            lag_name=None,
            default_vlan=1
        )

        mock_prepare.assert_called_once_with(
            self.api, topic=None, version="1.0"
        )
        mock_cctxt.call.assert_called_once_with(
            self.context,
            "update_port",
            switch_id="switch-01",
            port_name="port-01",
            description="Test port",
            mode="access",
            native_vlan=100,
            allowed_vlans=None,
            lag_name=None,
            default_vlan=1
        )
        self.assertEqual({"status": "configured"}, result)

    @mock.patch.object(rpcapi.NetworkingAPI, "_prepare_call", autospec=True)
    def test_update_port_with_allowed_vlans(self, mock_prepare):
        """Test update_port call with allowed VLANs."""
        mock_cctxt = mock.Mock()
        mock_prepare.return_value = mock_cctxt
        mock_cctxt.call.return_value = {"status": "configured"}

        result = self.api.update_port(
            self.context,
            "switch-01",
            "port-01",
            "Test port",
            "trunk",
            100,
            allowed_vlans=[101, 102],
            lag_name="po1",
            default_vlan=1
        )

        mock_prepare.assert_called_once_with(
            self.api, topic=None, version="1.0"
        )
        mock_cctxt.call.assert_called_once_with(
            self.context,
            "update_port",
            switch_id="switch-01",
            port_name="port-01",
            description="Test port",
            mode="trunk",
            native_vlan=100,
            allowed_vlans=[101, 102],
            lag_name="po1",
            default_vlan=1
        )
        self.assertEqual({"status": "configured"}, result)

    @mock.patch.object(rpcapi.NetworkingAPI, "_prepare_call", autospec=True)
    def test_update_port_with_custom_topic(self, mock_prepare):
        """Test update_port call with custom topic."""
        mock_cctxt = mock.Mock()
        mock_prepare.return_value = mock_cctxt
        mock_cctxt.call.return_value = {"status": "configured"}

        result = self.api.update_port(
            self.context,
            "switch-01",
            "port-01",
            "Test port",
            "access",
            100,
            default_vlan=1,
            topic="custom-topic",
        )

        mock_prepare.assert_called_once_with(
            self.api, topic="custom-topic", version="1.0"
        )
        mock_cctxt.call.assert_called_once_with(
            self.context,
            "update_port",
            switch_id="switch-01",
            port_name="port-01",
            description="Test port",
            mode="access",
            native_vlan=100,
            allowed_vlans=None,
            lag_name=None,
            default_vlan=1,
        )
        self.assertEqual({"status": "configured"}, result)

    @mock.patch.object(rpcapi.NetworkingAPI, "_prepare_call", autospec=True)
    def test_reset_port_success(self, mock_prepare):
        """Test successful reset_port call."""
        mock_cctxt = mock.Mock()
        mock_prepare.return_value = mock_cctxt
        mock_cctxt.call.return_value = {"status": "reset"}

        result = self.api.reset_port(
            self.context,
            "switch-01",
            "port-01",
            100,
            default_vlan=1)

        mock_prepare.assert_called_once_with(
            self.api, topic=None, version="1.0"
        )
        mock_cctxt.call.assert_called_once_with(
            self.context,
            "reset_port",
            switch_id="switch-01",
            port_name="port-01",
            native_vlan=100,
            allowed_vlans=None,
            default_vlan=1,
        )
        self.assertEqual({"status": "reset"}, result)

    @mock.patch.object(rpcapi.NetworkingAPI, "_prepare_call", autospec=True)
    def test_reset_port_with_allowed_vlans(self, mock_prepare):
        """Test reset_port call with allowed VLANs."""
        mock_cctxt = mock.Mock()
        mock_prepare.return_value = mock_cctxt
        mock_cctxt.call.return_value = {"status": "reset"}

        result = self.api.reset_port(
            self.context,
            "switch-01",
            "port-01",
            100,
            allowed_vlans=[101, 102],
            default_vlan=1
        )

        mock_prepare.assert_called_once_with(
            self.api, topic=None, version="1.0"
        )
        mock_cctxt.call.assert_called_once_with(
            self.context,
            "reset_port",
            switch_id="switch-01",
            port_name="port-01",
            native_vlan=100,
            allowed_vlans=[101, 102],
            default_vlan=1,
        )
        self.assertEqual({"status": "reset"}, result)

    @mock.patch.object(rpcapi.NetworkingAPI, "_prepare_call", autospec=True)
    def test_reset_port_with_custom_topic(self, mock_prepare):
        """Test reset_port call with custom topic."""
        mock_cctxt = mock.Mock()
        mock_prepare.return_value = mock_cctxt
        mock_cctxt.call.return_value = {"status": "reset"}

        result = self.api.reset_port(
            self.context,
            "switch-01",
            "port-01",
            100,
            default_vlan=1,
            topic="custom-topic"
        )

        mock_prepare.assert_called_once_with(
            self.api, topic="custom-topic", version="1.0"
        )
        mock_cctxt.call.assert_called_once_with(
            self.context,
            "reset_port",
            switch_id="switch-01",
            port_name="port-01",
            native_vlan=100,
            allowed_vlans=None,
            default_vlan=1,
        )
        self.assertEqual({"status": "reset"}, result)

    @mock.patch.object(rpcapi.NetworkingAPI, "_prepare_call", autospec=True)
    def test_get_switches_success(self, mock_prepare):
        """Test successful get_switches call."""
        mock_cctxt = mock.Mock()
        mock_prepare.return_value = mock_cctxt
        expected_switches = {
            "switch-01": {"name": "switch-01", "status": "connected"},
            "switch-02": {"name": "switch-02", "status": "connected"},
        }
        mock_cctxt.call.return_value = expected_switches

        result = self.api.get_switches(self.context)

        mock_prepare.assert_called_once_with(
            self.api, topic=None, version="1.0"
        )
        mock_cctxt.call.assert_called_once_with(self.context, "get_switches")
        self.assertEqual(expected_switches, result)

    @mock.patch.object(rpcapi.NetworkingAPI, "_prepare_call", autospec=True)
    def test_get_switches_with_custom_topic(self, mock_prepare):
        """Test get_switches call with custom topic."""
        mock_cctxt = mock.Mock()
        mock_prepare.return_value = mock_cctxt
        expected_switches = {}
        mock_cctxt.call.return_value = expected_switches

        result = self.api.get_switches(self.context, topic="custom-topic")

        mock_prepare.assert_called_once_with(
            self.api, topic="custom-topic", version="1.0"
        )
        mock_cctxt.call.assert_called_once_with(self.context, "get_switches")
        self.assertEqual(expected_switches, result)

    @mock.patch.object(rpcapi.NetworkingAPI, "_prepare_call", autospec=True)
    def test_update_lag_success(self, mock_prepare):
        """Test successful update_lag call."""
        mock_cctxt = mock.Mock()
        mock_prepare.return_value = mock_cctxt
        mock_cctxt.call.return_value = {"status": "configured"}

        result = self.api.update_lag(
            self.context,
            ["switch-01", "switch-02"],
            "lag-01",
            "Test LAG",
            "trunk",
            100,
            "lacp",
            allowed_vlans=[101, 102],
            default_vlan=1,
        )

        mock_prepare.assert_called_once_with(
            self.api, topic=None, version="1.0"
        )
        mock_cctxt.call.assert_called_once_with(
            self.context,
            "update_lag",
            switch_ids=["switch-01", "switch-02"],
            lag_name="lag-01",
            description="Test LAG",
            mode="trunk",
            native_vlan=100,
            aggregation_mode="lacp",
            allowed_vlans=[101, 102],
            default_vlan=1,
        )
        self.assertEqual({"status": "configured"}, result)

    @mock.patch.object(rpcapi.NetworkingAPI, "_prepare_call", autospec=True)
    def test_update_lag_without_allowed_vlans(self, mock_prepare):
        """Test update_lag call without allowed VLANs."""
        mock_cctxt = mock.Mock()
        mock_prepare.return_value = mock_cctxt
        mock_cctxt.call.return_value = {"status": "configured"}

        result = self.api.update_lag(
            self.context,
            ["switch-01"],
            "lag-01",
            "Test LAG",
            "access",
            100,
            "static",
            default_vlan=1,
        )

        mock_prepare.assert_called_once_with(
            self.api, topic=None, version="1.0"
        )
        mock_cctxt.call.assert_called_once_with(
            self.context,
            "update_lag",
            switch_ids=["switch-01"],
            lag_name="lag-01",
            description="Test LAG",
            mode="access",
            native_vlan=100,
            aggregation_mode="static",
            allowed_vlans=None,
            default_vlan=1,
        )
        self.assertEqual({"status": "configured"}, result)

    @mock.patch.object(rpcapi.NetworkingAPI, "_prepare_call", autospec=True)
    def test_update_lag_with_custom_topic(self, mock_prepare):
        """Test update_lag call with custom topic."""
        mock_cctxt = mock.Mock()
        mock_prepare.return_value = mock_cctxt
        mock_cctxt.call.return_value = {"status": "configured"}

        result = self.api.update_lag(
            self.context,
            ["switch-01"],
            "lag-01",
            "Test LAG",
            "access",
            100,
            "static",
            default_vlan=1,
            topic="custom-topic",
        )

        mock_prepare.assert_called_once_with(
            self.api, topic="custom-topic", version="1.0"
        )
        mock_cctxt.call.assert_called_once_with(
            self.context,
            "update_lag",
            switch_ids=["switch-01"],
            lag_name="lag-01",
            description="Test LAG",
            mode="access",
            native_vlan=100,
            aggregation_mode="static",
            allowed_vlans=None,
            default_vlan=1,
        )
        self.assertEqual({"status": "configured"}, result)

    @mock.patch.object(rpcapi.NetworkingAPI, "_prepare_call", autospec=True)
    def test_delete_lag_success(self, mock_prepare):
        """Test successful delete_lag call."""
        mock_cctxt = mock.Mock()
        mock_prepare.return_value = mock_cctxt
        mock_cctxt.call.return_value = {"status": "deleted"}

        result = self.api.delete_lag(
            self.context, ["switch-01", "switch-02"], "lag-01"
        )

        mock_prepare.assert_called_once_with(
            self.api, topic=None, version="1.0"
        )
        mock_cctxt.call.assert_called_once_with(
            self.context,
            "delete_lag",
            switch_ids=["switch-01", "switch-02"],
            lag_name="lag-01",
        )
        self.assertEqual({"status": "deleted"}, result)

    @mock.patch.object(rpcapi.NetworkingAPI, "_prepare_call", autospec=True)
    def test_delete_lag_with_custom_topic(self, mock_prepare):
        """Test delete_lag call with custom topic."""
        mock_cctxt = mock.Mock()
        mock_prepare.return_value = mock_cctxt
        mock_cctxt.call.return_value = {"status": "deleted"}

        result = self.api.delete_lag(
            self.context, ["switch-01"], "lag-01", topic="custom-topic"
        )

        mock_prepare.assert_called_once_with(
            self.api, topic="custom-topic", version="1.0"
        )
        mock_cctxt.call.assert_called_once_with(
            self.context,
            "delete_lag",
            switch_ids=["switch-01"],
            lag_name="lag-01",
        )
        self.assertEqual({"status": "deleted"}, result)


class TestNetworkingAPIVersionCap(test_base.TestCase):
    """Test cases for NetworkingAPI version cap handling."""

    def setUp(self):
        super(TestNetworkingAPIVersionCap, self).setUp()
        self.context = mock.Mock()

    @mock.patch.object(rpc, "NETWORKING_TOPIC", "test-topic")
    def test_version_cap_from_release_mapping(self):
        """Test version cap is set from release mapping."""
        with mock.patch.object(
            rpcapi.versions,
            "RELEASE_MAPPING",
            {"zed": {"rpc": "1.0"}}):
            CONF.set_override("pin_release_version", "zed")
            api = rpcapi.NetworkingAPI()
            # Version cap should be applied from release mapping
            self.assertIsNotNone(api.client)

    @mock.patch.object(rpc, "NETWORKING_TOPIC", "test-topic")
    def test_version_cap_fallback_to_current(self):
        """Test version cap falls back to current version."""
        with mock.patch.object(rpcapi.versions, "RELEASE_MAPPING", {}):
            CONF.set_override("pin_release_version", None)
            api = rpcapi.NetworkingAPI()
            # Should use current RPC_API_VERSION
            self.assertIsNotNone(api.client)

    @mock.patch.object(rpc, "NETWORKING_TOPIC", "test-topic")
    def test_version_cap_no_pin_release_version(self):
        """Test version cap when pin_release_version is not set."""
        with mock.patch.object(
            rpcapi.versions,
            "RELEASE_MAPPING",
            {"zed": {"rpc": "1.0"}}
        ):
            CONF.set_override("pin_release_version", None)
            api = rpcapi.NetworkingAPI()
            # Should use current RPC_API_VERSION
            self.assertIsNotNone(api.client)



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
"""
Tests for pluggable API middleware loaded via stevedore entry points.

This demonstrates how external packages can register custom WSGI middleware
via the ironic.api.middleware entry point namespace.
"""

import json
import re
import unittest
from unittest import mock

from oslo_config import cfg
import webob
import webob.dec

from ironic.api import app
from ironic.conf import CONF


# Example middleware: Portgroup name validation
# This is a simplified version of what an external package might provide
# to validate portgroup names match a required format.

PORTGROUP_NAME_PATTERN = re.compile(r"^.+-port-channel(\d+)$")
PORT_CHANNEL_MIN = 100
PORT_CHANNEL_MAX = 998


class PortgroupNameValidationMiddleware:
    """Example WSGI middleware that validates portgroup names.

    This demonstrates a real-world use case: validating that portgroup
    names match a required format for integration with external switch
    configuration systems.

    Intercepts POST /v1/portgroups requests and validates that names
    match the format: {node_name}-port-channel{number}
    """

    def __init__(self, application):
        self.application = application

    @webob.dec.wsgify
    def __call__(self, req):
        # Only validate portgroup create requests
        if (req.method == "POST"
                and req.path_info.rstrip("/") == "/v1/portgroups"):
            return self._validate_create(req)
        return req.get_response(self.application)

    def _validate_create(self, req):
        """Validate portgroup name on create."""
        try:
            body = json.loads(req.body)
        except (json.JSONDecodeError, ValueError):
            # Let Ironic handle malformed JSON
            return req.get_response(self.application)

        name = body.get("name")
        if not name:
            return self._error_response(
                "Portgroup name is required and cannot be empty."
            )

        match = PORTGROUP_NAME_PATTERN.match(name)
        if not match:
            return self._error_response(
                f"Portgroup name '{name}' must match format "
                "'{node_name}-port-channel{{number}}'"
            )

        port_channel_num = int(match.group(1))
        if not (PORT_CHANNEL_MIN <= port_channel_num <= PORT_CHANNEL_MAX):
            return self._error_response(
                f"Portgroup name '{name}' has invalid port-channel number "
                f"{port_channel_num}. Must be between {PORT_CHANNEL_MIN} "
                f"and {PORT_CHANNEL_MAX}."
            )

        return req.get_response(self.application)

    def _error_response(self, message):
        """Return a 400 Bad Request response."""
        error_body = {
            "error_message": json.dumps({
                "faultstring": message,
                "faultcode": "Client",
                "debuginfo": None
            })
        }
        return webob.Response(
            status=400,
            content_type="application/json",
            body=json.dumps(error_body).encode("utf-8"),
        )


class TestLoadCustomMiddleware(unittest.TestCase):
    """Tests for the _load_custom_middleware function."""

    def test_load_no_middleware(self):
        """Test that no middleware is loaded when config is empty."""
        test_app = mock.MagicMock()
        result = app._load_custom_middleware(test_app, [])
        # With empty list, the original app should be returned unchanged
        self.assertEqual(test_app, result)

    @mock.patch('stevedore.NamedExtensionManager', autospec=True)
    def test_load_portgroup_validation_middleware(self, mock_manager):
        """Test loading portgroup name validation middleware."""
        test_app = mock.MagicMock()

        # Create a mock extension for portgroup validation
        mock_ext = mock.MagicMock()
        mock_ext.name = 'portgroup-name-validation'
        mock_ext.plugin = PortgroupNameValidationMiddleware

        mock_manager.return_value = [mock_ext]

        result = app._load_custom_middleware(
            test_app, ['portgroup-name-validation']
        )

        # Verify stevedore was called correctly
        mock_manager.assert_called_once_with(
            'ironic.api.middleware',
            names=['portgroup-name-validation'],
            invoke_on_load=False,
            on_missing_entrypoints_callback=app._missing_middleware_callback,
            name_order=True,
        )

        # Verify the middleware was applied
        self.assertIsInstance(result, PortgroupNameValidationMiddleware)
        self.assertEqual(result.application, test_app)

    @mock.patch('stevedore.NamedExtensionManager', autospec=True)
    def test_load_multiple_middleware(self, mock_manager):
        """Test loading multiple middleware in order."""
        test_app = mock.MagicMock()

        # Create mock extensions
        mock_ext1 = mock.MagicMock()
        mock_ext1.name = 'portgroup-name-validation'
        mock_ext1.plugin = PortgroupNameValidationMiddleware

        mock_ext2 = mock.MagicMock()
        mock_ext2.name = 'another-middleware'
        mock_ext2.plugin = PortgroupNameValidationMiddleware

        mock_manager.return_value = [mock_ext1, mock_ext2]

        result = app._load_custom_middleware(
            test_app, ['portgroup-name-validation', 'another-middleware']
        )

        # The result should be middleware2 wrapping middleware1 wrapping app
        self.assertIsInstance(result, PortgroupNameValidationMiddleware)
        self.assertIsInstance(
            result.application, PortgroupNameValidationMiddleware
        )
        self.assertEqual(result.application.application, test_app)


class TestMissingMiddlewareCallback(unittest.TestCase):
    """Tests for the _missing_middleware_callback function."""

    def test_missing_middleware_raises_runtime_error(self):
        """Test that missing middleware raises RuntimeError."""
        missing_names = ['missing1', 'missing2']
        self.assertRaises(
            RuntimeError,
            app._missing_middleware_callback,
            missing_names
        )

    def test_missing_middleware_error_message(self):
        """Test that error message contains missing middleware names."""
        missing_names = ['portgroup-validation', 'rate-limiter']
        try:
            app._missing_middleware_callback(missing_names)
            self.fail("Expected RuntimeError to be raised")
        except RuntimeError as e:
            self.assertIn('portgroup-validation', str(e))
            self.assertIn('rate-limiter', str(e))


class TestMiddlewareConfigOption(unittest.TestCase):
    """Tests for the [api] middleware config option."""

    def setUp(self):
        # Register the api config options if not already registered
        from ironic.conf import api as api_conf
        try:
            api_conf.register_opts(CONF)
        except cfg.DuplicateOptError:
            pass  # Already registered

    def test_middleware_config_default_empty(self):
        """Test that middleware config defaults to empty list."""
        self.assertEqual([], CONF.api.middleware)

    def test_middleware_config_accepts_list(self):
        """Test that middleware config accepts a list of names."""
        CONF.set_override(
            'middleware',
            ['portgroup-name-validation', 'rate-limiter'],
            group='api'
        )
        try:
            self.assertEqual(
                ['portgroup-name-validation', 'rate-limiter'],
                CONF.api.middleware
            )
        finally:
            CONF.clear_override('middleware', group='api')


class TestPortgroupNameValidationMiddleware(unittest.TestCase):
    """Tests for the example portgroup name validation middleware.

    These tests demonstrate how the middleware validates portgroup names
    to ensure they match the required format for undersync integration.
    """

    def setUp(self):
        # Create a simple WSGI app that returns 200 OK
        def simple_app(environ, start_response):
            start_response('200 OK', [('Content-Type', 'application/json')])
            return [b'{"result": "ok"}']
        self.middleware = PortgroupNameValidationMiddleware(simple_app)

    def _make_request(self, method, path, body=None):
        """Create a webob Request for testing."""
        req = webob.Request.blank(path)
        req.method = method
        if body:
            req.body = json.dumps(body).encode('utf-8')
            req.content_type = 'application/json'
        return req

    def test_valid_portgroup_name(self):
        """Test that valid portgroup names are accepted."""
        req = self._make_request(
            'POST', '/v1/portgroups',
            {'name': 'server01-port-channel100', 'node_uuid': 'test-uuid'}
        )
        resp = req.get_response(self.middleware)
        # Should pass through to the app (returns 200)
        self.assertEqual(200, resp.status_int)

    def test_invalid_portgroup_name_format(self):
        """Test that invalid portgroup name format is rejected."""
        req = self._make_request(
            'POST', '/v1/portgroups',
            {'name': 'invalid-name', 'node_uuid': 'test-uuid'}
        )
        resp = req.get_response(self.middleware)
        self.assertEqual(400, resp.status_int)
        body = json.loads(resp.body)
        error = json.loads(body['error_message'])
        self.assertIn('must match format', error['faultstring'])

    def test_request_passes_through(self):
        """Test that non-portgroup requests pass through unchanged."""
        req = self._make_request('GET', '/v1/nodes')
        resp = req.get_response(self.middleware)
        self.assertEqual(200, resp.status_int)

    def test_portgroup_get_request_passes_through(self):
        """Test that GET requests to portgroups pass through."""
        req = self._make_request('GET', '/v1/portgroups')
        resp = req.get_response(self.middleware)
        self.assertEqual(200, resp.status_int)



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


from oslo_config import cfg
from oslo_log import log

from ironic.common import rpc_service
from ironic.networking import utils as networking_utils

LOG = log.getLogger(__name__)
CONF = cfg.CONF


class NetworkingRPCService(rpc_service.BaseRPCService):
    """RPC service for the Ironic Networking Manager."""

    def __init__(self, host, manager_module, manager_class):
        super().__init__(host, manager_module, manager_class)
        self.graceful_shutdown = False

    def _rpc_transport(self):
        return networking_utils.rpc_transport()

    def _real_start(self):
        """Start the networking service."""
        super()._real_start()
        LOG.info(
            "Started networking RPC server for service %(service)s on "
            "host %(host)s.",
            {"service": self.topic, "host": self.host},
        )

    def stop(self):
        """Stop the networking service."""
        LOG.info(
            "Stopping networking RPC server for service %(service)s on "
            "host %(host)s.",
            {"service": self.topic, "host": self.host},
        )

        try:
            if hasattr(self.manager, "del_host"):
                self.manager.del_host()
        except Exception as e:
            LOG.exception(
                "Service error occurred when cleaning up "
                "the networking RPC manager. Error: %s",
                e,
            )

        try:
            if self.rpcserver is not None:
                self.rpcserver.stop()
                self.rpcserver.wait()
        except Exception as e:
            LOG.exception(
                "Service error occurred when stopping the "
                "networking RPC server. Error: %s",
                e,
            )

        super().stop(graceful=True)
        LOG.info(
            "Stopped networking RPC server for service %(service)s on "
            "host %(host)s.",
            {"service": self.topic, "host": self.host},
        )



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import ironic.common.trait_based_networking.base as tbn_base
import ironic.common.trait_based_networking.config_file as cf

from ironic.tests import base

from dataclasses import dataclass
import tempfile


EXAMPLE_CONFIG_FILE_LOCATION = "etc/ironic/trait_based_networks.yaml.sample"


class TraitBasedNetworkingConfigFileTestCase(base.TestCase):
    def setUp(self):
        super(TraitBasedNetworkingConfigFileTestCase, self).setUp()
        self.tmpdir = tempfile.TemporaryDirectory()

        self.addTypeEqualityFunc(
            tbn_base.NetworkTrait,
            lambda first, second, msg=None: first == second
        )

    def test_load_example_config_file(self):
        config_file = cf.ConfigFile(EXAMPLE_CONFIG_FILE_LOCATION)
        self.assertIsNotNone(config_file)
        self.assertTrue(config_file.validate()[0])
        config_file.parse()

    def test_validate(self):
        @dataclass
        class SubTestCase(object):
            description: str
            contents: str
            expected_valid: bool
            expected_reasons: list[str]

        subtests = [
            SubTestCase(
                "Valid - single trait",
                ("CUSTOM_TRAIT_NAME:\n"
                 "  order: 1\n"
                 "  actions:\n"
                 "    - action: bond_ports\n"
                 "      filter: port.vendor == 'vendor_string'\n"
                 "      min_count: 2\n"),
                True,
                [],
            ),
            SubTestCase(
                "Valid - Several traits",
                ("CUSTOM_TRAIT_NAME:\n"
                 "  actions:\n"
                 "    - action: bond_ports\n"
                 "      filter: port.vendor == 'vendor_string'\n"
                 "      min_count: 2\n"
                 "CUSTOM_TRAIT_2:\n"
                 "  actions:\n"
                 "    - action: attach_port\n"
                 "      filter: port.vendor != 'vendor_string'\n"
                 "      max_count: 2\n"
                 "CUSTOM_TRAIT_3:\n"
                 "  actions:\n"
                 "    - action: attach_port\n"
                 "      filter: port.vendor != 'vendor_string'\n"
                 "      max_count: 2\n"),
                True,
                [],
            ),
            SubTestCase(
                "Invalid - Missing trait has required entry missing",
                ("trait_name:\n"
                 "  actions:\n"
                 "  - action: bond_ports\n"),
                False,
                ["'trait_name' trait is missing 'filter' key"],
            ),
            SubTestCase(
                "Invalid - Unrecognized trait entry",
                ("CUSTOM_TRAIT_NAME:\n"
                 "  actions:\n"
                 "    - action: bond_ports\n"
                 "      filter: port.vendor == 'vendor_string'\n"
                 "      min_count: 2\n"
                 "      wrong: hi\n"),
                False,
                [("'CUSTOM_TRAIT_NAME' trait action has unrecognized key "
                 "'wrong'")],
            ),
            SubTestCase(
                "Invalid - Unrecognized action",
                ("CUSTOM_TRAIT_NAME:\n"
                 "  actions:\n"
                 "    - action: invalid\n"
                 "      filter: port.vendor == 'vendor_string'\n"
                 "      min_count: 2\n"),
                False,
                ["'CUSTOM_TRAIT_NAME' trait action has unrecognized action "
                 "'invalid'"],
            ),
            SubTestCase(
                ("Invalid - trait actions does not consist of a list of "
                 "actions"),
                ("CUSTOM_TRAIT_NAME:\n"
                 "  actions:\n"
                 "    action: bond_ports\n"
                 "    filter: port.vendor == 'vendor_string'\n"
                 "    min_count: 2\n"),
                False,
                [("'CUSTOM_TRAIT_NAME.actions' does not consist of a list "
                  "of actions")],
            ),
            SubTestCase(
                "Invalid - trait action has malformed filter expression",
                ("CUSTOM_TRAIT_NAME:\n"
                 "  actions:\n"
                 "    - action: bond_ports\n"
                 "      filter: port.vendor &= 'vendor_string'\n"
                 "      min_count: 2\n"),
                False,
                [("'CUSTOM_TRAIT_NAME' trait action has malformed filter "
                  "expression: 'port.vendor &= 'vendor_string''")],
            ),
            SubTestCase(
                "Invalid - several things wrong",
                ("CUSTOM_TRAIT_NAME:\n"
                 "  actions:\n"
                 "    - filter: port.vendor &= 'vendor_string'\n"
                 "      min_count: 2\n"
                 "      wrong: oops\n"),
                False,
                [("'CUSTOM_TRAIT_NAME' trait action has malformed filter "
                  "expression: 'port.vendor &= 'vendor_string''"),
                 "'CUSTOM_TRAIT_NAME' trait is missing 'action' key",
                 ("'CUSTOM_TRAIT_NAME' trait action has unrecognized key "
                  "'wrong'")],
            ),
            SubTestCase(
                "Invalid - group_and_attach_ports is invalid",
                ("CUSTOM_TRAIT_NAME:\n"
                 "  actions:\n"
                 "    - action: group_and_attach_ports\n"
                 "      filter: port.vendor == 'vendor_string'\n"
                 "      min_count: 0\n"),
                False,
                [("'CUSTOM_TRAIT_NAME' has an invalid "
                  "'group_and_attach_ports': group_and_attach_ports must have "
                  "a min_count of "
                  f"{tbn_base.DEFAULT_GROUP_AND_ATTACH_MIN_COUNT} "
                  "or greater. Got '0'.")],
            ),
        ]

        for subtest in subtests:
            with self.subTest(subtest=subtest):
                with tempfile.NamedTemporaryFile(
                        mode='w',
                        dir=self.tmpdir.name,
                        delete=False) as tmpfile:
                    tmpfile.write(subtest.contents)
                    tmpfile.close()
                    config_file = cf.ConfigFile(tmpfile.name)
                    valid, reasons = config_file.validate()
                    self.assertEqual(subtest.expected_valid, valid)
                    self.assertCountEqual(subtest.expected_reasons, reasons)

    def test_parse(self):
        contents = (
            "CUSTOM_TRAIT_NAME:\n"
            "  actions:\n"
            "    - action: bond_ports\n"
            "      filter: port.vendor == 'vendor_string'\n"
            "      min_count: 2\n")

        with tempfile.NamedTemporaryFile(
                mode='w',
                dir=self.tmpdir.name,
                delete=False) as tmpfile:
            tmpfile.write(contents)
            tmpfile.close()
            config_file = cf.ConfigFile(tmpfile.name)
            valid, reasons = config_file.validate()
            self.assertTrue(valid)
            self.assertCountEqual(reasons, [])

            config_file.parse()
            result = config_file.traits()

            self.assertCountEqual(result, [
                tbn_base.NetworkTrait(
                    "CUSTOM_TRAIT_NAME",
                    [tbn_base.TraitAction(
                        "CUSTOM_TRAIT_NAME",
                        tbn_base.Actions("bond_ports"),
                        tbn_base.FilterExpression.parse(
                            "port.vendor == 'vendor_string'"),
                        min_count=2)]
                )
            ])
